/*******************************************************************************
** sdla_remora.h	
**
** Author: 	Alex Feldman  <al.feldman@sangoma.com>
**
** Copyright:	(c) 2005 Sangoma Technologies Inc.
**
**		This program is free software; you can redistribute it and/or
**		modify it under the terms of the GNU General Public License
**		as published by the Free Software Foundation; either version
**		2 of the License, or (at your option) any later version.
** ============================================================================
** Oct 6, 2005	Alex Feldman	Initial version.
*******************************************************************************/

#ifndef __SDLA_REMORA_H
# define __SDLA_REMORA_H

#ifdef __SDLA_REMORA_SRC
# define EXTERN
#else
# define EXTERN extern
#endif

#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
# include <net/sdla_remora_proslic.h>
#else
# include <linux/sdla_remora_proslic.h>
#endif

/*******************************************************************************
**			  DEFINES and MACROS
*******************************************************************************/

#define IS_FXOFXS_CARD(card)		IS_FXOFXS_FEMEDIA(&(card)->fe)

#define MAX_REMORA_MODULES		24
#define MAX_FXOFXS_CHANNELS		MAX_REMORA_MODULES

#define MOD_TYPE_NONE			0
#define MOD_TYPE_FXS			1
#define MOD_TYPE_FXO			2
#define WP_REMORA_DECODE_TYPE(type)			\
		(type == MOD_TYPE_FXS) ? "FXS" :	\
		(type == MOD_TYPE_FXO) ? "FXO" :	\
					"Unknown"

/* SPI interface */
#define MOD_CHAIN_DISABLED	0
#define MOD_CHAIN_ENABLED	1

#define SPI_INTERFACE_REG	0x54

#define MOD_SPI_ADDR_FXS_READ	0x80
#define MOD_SPI_ADDR_FXS_WRITE	0x00

#define MOD_SPI_CS_FXS_CHIP_0	0x01
#define MOD_SPI_CS_FXS_CHIP_1	0x02

#define MOD_SPI_CS_FXO_CHIP_1	0x08
#define MOD_SPI_CS_FXO_WRITE	0x00
#define MOD_SPI_CS_FXO_READ	0x40

#define MOD_SPI_CTRL_V3_START	0x08	/* bit 27 used as START in version 1-2-3 */
#define MOD_SPI_CTRL_START	0x80
#define MOD_SPI_CTRL_FXS	0x10
#define MOD_SPI_CTRL_CHAIN	0x20

#define MOD_SPI_RESET		0x40000000
#define MOD_SPI_BUSY		0x80000000
#define MOD_SPI_V3_STAT		0x08000000	/* bit 27 used as START in version 1-2-3 */
#define MOD_SPI_START		0x08000000


#define WAN_FXS_NUM_REGS		109
#define WAN_FXS_NUM_INDIRECT_REGS	105

#define WAN_FXO_NUM_REGS		59

#define WAN_REMORA_READREG		0x01;

#define RM_FE_MAGIC	0x150D

/* Front-End UDP command */
#define WAN_FE_TONES	(WAN_FE_UDP_CMD_START + 0)
#define WAN_FE_RING	(WAN_FE_UDP_CMD_START + 1)
#define WAN_FE_REGDUMP	(WAN_FE_UDP_CMD_START + 2)


/* Interrup types */
#define WAN_RM_INTR_NONE	0x00
#define WAN_RM_INTR_FXS_DTMF	0x01

/*******************************************************************************
**			  TYPEDEF STRUCTURE
*******************************************************************************/
typedef struct sdla_remora_cfg_ {
	int	not_used;
	int	fxo_opermode;
/*	int	tdmv_law;*/	/* WAN_TDMV_ALAW or WAN_TDMV_MULAW */
	int	reversepolarity;
} sdla_remora_cfg_t;

typedef struct {
	unsigned char direct[WAN_FXS_NUM_REGS];
	unsigned short indirect[WAN_FXS_NUM_INDIRECT_REGS];
} wan_remora_fxs_regs_t;

typedef struct {
	unsigned char direct[WAN_FXO_NUM_REGS];
} wan_remora_fxo_regs_t;

typedef struct {
	int	mod_no;
	int	type;
	
	union {
		wan_remora_fxs_regs_t	regs_fxs;
		wan_remora_fxo_regs_t	regs_fxo;
	} u;
} wan_remora_udp_t;

#if defined(WAN_KERNEL)

#define NUM_CAL_REGS		12

#define __WRITE_RM_REG(mod_no,type,chain,reg,val)			\
	fe->write_fe_reg(fe->card,(int)mod_no,(int)type,(int)chain,(int)reg,(int)val)
#define __READ_RM_REG(mod_no,type,chain,reg)				\
	fe->read_fe_reg(fe->card, (int)mod_no,(int)type,(int)chain,(int)reg)

#define WRITE_RM_REG(mod_no,reg,val)					\
	fe->write_fe_reg(	fe->card,				\
				(int)mod_no,				\
				fe->fe_param.remora.mod[mod_no].type,	\
				fe->fe_param.remora.mod[mod_no].chain,	\
				(int)reg, (int)val)
#define READ_RM_REG(mod_no,reg)						\
	fe->read_fe_reg(	fe->card,				\
				(int)mod_no,				\
				fe->fe_param.remora.mod[mod_no].type,	\
				fe->fe_param.remora.mod[mod_no].chain,	\
				(int)reg)

enum proslic_power_warn {
	PROSLIC_POWER_UNKNOWN = 0,
	PROSLIC_POWER_ON,
	PROSLIC_POWER_WARNED,
};

typedef struct {
	unsigned char vals[NUM_CAL_REGS];
} callregs_t;

typedef struct {
	int	ready;
	
} wp_remora_fxo_t;

typedef struct {
	int		ready;
	int		idletxhookstate;
	int		oldrxhook;
	enum proslic_power_warn proslic_power;
	callregs_t	callregs;
} wp_remora_fxs_t;

typedef struct {
	int	type;
	int	chain;
	union {
		wp_remora_fxo_t	fxo;
		wp_remora_fxs_t	fxs;
	} u;

} wp_remora_module_t;

typedef struct sdla_remora_param {
	int			not_used;

	wp_remora_module_t	mod[MAX_REMORA_MODULES];

	unsigned long		module_map;		/* Map of available module */
	int			max_fe_channels;	/* Number of available modules */
	
	unsigned char		critical;
	wan_timer_t		timer;
	unsigned char		timer_cmd;
	int			udp_mod_no;

} sdla_remora_param_t;


#endif /* WAN_KERNEL */

/*******************************************************************************
**			  FUNCTION PROTOTYPES
*******************************************************************************/
extern int	wp_remora_iface_init(void*);

#undef EXTERN
#endif	/* __SDLA_REMORA_H */
