/*
 * Copyright (c) 1997, 1998, 1999
 *	Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Alex Feldman.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Alex Feldman AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Alex Feldman OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: sdla_pci.c,v 1.14 2006/02/20 14:39:34 sangoma Exp $
 */

/*****************************************************************************
 * sdla_pci.c	SDLA PCI Support Module.  Main module.
 *
 *		This module is a compeled with kernel sources.
 *
 * Author:	Alex Feldman	<al.feldman@sangoma.com>
 ****************************************************************************
 */

#undef SDLA_PCI_DEBUG 	/* uncomment this for debug messages */
#undef INTR_TEST	/* uncomment this to check that driver RX INTR */
#define WAN_KERNEL

/*#include "sdla.h"*/		/* number of Sangoma cards */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/syslog.h>
#include <sys/errno.h>
#include <sys/mbuf.h>
#include <sys/proc.h>
#include <sys/sockio.h>
#include <sys/socket.h>
#include <sys/kernel.h>
#include <sys/device.h>
#include <machine/bus.h>

#include <net/wanpipe_debug.h>
#include <net/sdlapci.h>
#include <net/sdlasfm.h>
#include <net/sdla_dev.h>	/* special header for S508 board */

/* Read PCI SUBVENDOR ID */
#define PCI_SUBVENDOR_MASK	0xFFFF
#define PCI_SUBVENDOR(pa)				\
	pci_conf_read(pa->pa_pc, pa->pa_tag, PCI_SUBSYS_ID_REG) & PCI_SUBVENDOR_MASK
#define PCI_SUBSYSTEM(pa)				\
	pci_conf_read(pa->pa_pc, pa->pa_tag, PCI_SUBSYS_ID_REG) >> 16
#define PCI_DEVICE_MASK		0xFFFF0000
#define PCI_DEVICE(id)		((id & PCI_DEVICE_MASK ) >> 16)

/* SDLA PCI device relative entry point 				*/
int 	sdla_pci_match	__P((struct device*, void*, void*));
void 	sdla_pci_attach __P((struct device*, struct device*, void*));
int 	sdla_pci_intr(void*);

struct cfdriver sdla_pci_cd = {
	NULL, "sdla_pci", DV_IFNET
};

struct cfattach sdla_pci_ca = {
	sizeof(struct sdla_softc), sdla_pci_match, sdla_pci_attach
};

/************************************************************************/

/* For debugging convenience */
#ifdef SDLA_DEV_DEBUG
 #define dlog(x) log x
#else
 #define dlog(x)
#endif /* SDLA_DEV_DEBUG */

/******* PCI Device Driver Entry Points ********************************/
static int sdla_supported(struct pci_attach_args *pa)
{
	u_int16_t	vendor = PCI_VENDOR(pa->pa_id);	
	u_int16_t	subvendor = PCI_SUBVENDOR(pa);	
	u_int16_t	device = PCI_DEVICE(pa->pa_id);	

	switch(vendor){
	case V3_VENDOR_ID:
		if (subvendor == SANGOMA_SUBSYS_VENDOR){
			return 0;
		}
		break;
	case PCI_VENDOR_ID_GSI:
		if (device == PCI_DEVICE_ID_GSI_PULSAR){
			return 0;
		}
		break;
	case SANGOMA_PCI_VENDOR_OLD:
		if (device == SANGOMA_PCI_DEVICE){
			return 0;
		}
		break;
	case SANGOMA_PCI_VENDOR:
		switch(device){
		case SANGOMA_PCI_DEVICE:
		case SANGOMA_PCI_4_DEVICE:
			return 0;
		default:
			return 0;	/* all new cards */
		}
		break;
	}
	return -EINVAL;
}

int sdla_pci_match(parent, match, aux)
	struct device* parent;
	void* match, *aux;
{
	struct pci_attach_args*	pa = aux;
	int			adapter_type = 0x00;

	if (sdla_supported(pa)){
		return 0;
	}

	adapter_type = pci_conf_read(pa->pa_pc, pa->pa_tag, PCI_SUBSYS_ID_WORD);
	if (!sdla_probe(adapter_type)){
		return 0;
	}
	return 1;
}

#define PCI_CBIO	0x10
void sdla_pci_attach(parent, self, aux)
	struct device *parent, *self;
	void*	aux;
{
	struct sdla_softc	*sc = (struct sdla_softc*)self;
	sdladev_t		*pci_cfg = NULL, *pci_cfg_nxt = NULL;
	struct pci_attach_args	*pa = aux;
	pci_chipset_tag_t	pc = pa->pa_pc;
	pci_intr_handle_t	ih;
	u_int16_t		vendor_id = PCI_VENDOR(pa->pa_id);	
	u_int16_t		subvendor_id = PCI_SUBVENDOR(pa);	
	u_int16_t		subsystem_id = PCI_SUBSYSTEM(pa);	
	u_int16_t		device_id = PCI_DEVICE(pa->pa_id);	
	const char*		intrstr = NULL;
	int 			adapter_type = 0x00,
				adapter_subtype = 0x00,
				dual_cpu = 0, i = 0, line_no = 1, type = 0;

	adapter_type = PCI_PRODUCT(
		pci_conf_read(pa->pa_pc, pa->pa_tag, PCI_SUBSYS_ID_REG));
	switch(vendor_id){
	case V3_VENDOR_ID:
		type = SDLA_S514;
		if (adapter_type == S5148_ADPTR_1_CPU_T1E1){
			adapter_type = S5144_ADPTR_1_CPU_T1E1;
		}
		switch(adapter_type){
		case S5142_ADPTR_2_CPU_SERIAL:
		case S5147_ADPTR_2_CPU_T1E1:
			dual_cpu = 1;
			break;
		}
		break;
	case PCI_VENDOR_ID_GSI:
		type = SDLA_ADSL;
		adapter_type = S518_ADPTR_1_CPU_ADSL;
		break;
	case SANGOMA_PCI_VENDOR_OLD:
	case SANGOMA_PCI_VENDOR:
		type = SDLA_AFT;
		switch(subvendor_id){
		case A101_1TE1_SUBSYS_VENDOR:
			adapter_type	= A101_ADPTR_1TE1;
			break;
		case A101_2TE1_SUBSYS_VENDOR:
			adapter_type	= A101_ADPTR_2TE1;
			dual_cpu = 1;
			break;
		case A104_4TE1_SUBSYS_VENDOR:
			adapter_type	= A104_ADPTR_4TE1;
			line_no = 4;
			break;
		case A104_1TE1_SHARK_SUBSYS_VENDOR:
			adapter_type	= A101_ADPTR_1TE1;
			adapter_subtype	= AFT_SUBTYPE_SHARK;
			break;
		case A104_2TE1_SHARK_SUBSYS_VENDOR:
			adapter_type	= A101_ADPTR_2TE1;
			adapter_subtype	= AFT_SUBTYPE_SHARK;
			line_no = 2;
			break;
		case A104_4TE1_SHARK_SUBSYS_VENDOR:
			adapter_type	= A104_ADPTR_4TE1;
			adapter_subtype	= AFT_SUBTYPE_SHARK;
			line_no = 4;
			break;
		case A104_8TE1_SHARK_SUBSYS_VENDOR:
			adapter_type	= A108_ADPTR_8TE1;
			adapter_subtype	= AFT_SUBTYPE_SHARK;
			line_no = 8;
			break;
		case A200_REMORA_SHARK_SUBSYS_VENDOR:
			adapter_type	= A200_ADPTR_ANALOG;
			adapter_subtype	= AFT_SUBTYPE_SHARK;
			break;
		default:
			return;
		}
		break;
	default:
		return;
	}

	sc->pa		= *pa;
	/* Map and establish the interrupt */
	if (pci_intr_map(pa,&ih)){
		printf(": couldn't map interrupt\n");
		return;
	}
	intrstr = pci_intr_string(pc, ih);

	/* For CPU A */
	pci_cfg			= sdla_attach(sc, adapter_type);
	pci_cfg->type 		= type;
	pci_cfg->adapter_type	= adapter_type;
	pci_cfg->adapter_subtype= adapter_subtype;
	pci_cfg->vendor_id	= vendor_id;
	pci_cfg->subvendor_id	= subvendor_id;
	pci_cfg->subsystem_id	= subsystem_id;
	pci_cfg->device_id	= device_id;
	memset(pci_cfg->intr_arg, 0, sizeof(sdla_intr_arg_t)*SDLA_MAX_PORTS);
	pci_cfg->irq		= pa->pa_intrline;
	pci_cfg->u.pci.cpu_no	= SDLA_CPU_A;
	pci_cfg->u.pci.slot_no	= pa->pa_device;
	pci_cfg->u.pci.bus_no	= pa->pa_bus;
	for(i=0;i<line_no;i++){
		sc->sc_ih = pci_intr_establish(
				pc, ih, IPL_NET,
				sdla_pci_intr,
				&pci_cfg->intr_arg[i],
				"sdla");
		if (sc->sc_ih == NULL){
			printf(": couldn't establish interrupt for cpu A");
			if (intrstr != NULL){
				printf(" at %s", intrstr);
			}
			printf("\n");
			return;
		}
		pci_cfg->intr_arg[i].sd		= (void*)pci_cfg;
		pci_cfg->intr_arg[i].line_no	= i;
		pci_cfg->intr_arg[i].ready	= 1;
	}

	if (dual_cpu){
	    	/* Map interrupt handler for second CPU */
	    	pci_cfg_nxt			= sdla_attach(sc, adapter_type);
	    	pci_cfg_nxt->type 		= type;
		pci_cfg_nxt->adapter_type	= adapter_type;
		pci_cfg_nxt->adapter_subtype	= adapter_subtype;
		pci_cfg_nxt->vendor_id		= vendor_id;
		pci_cfg_nxt->subvendor_id	= subvendor_id;
		pci_cfg_nxt->subsystem_id	= subsystem_id;
		pci_cfg_nxt->device_id		= device_id;
		memset(pci_cfg_nxt->intr_arg, 0,
				sizeof(sdla_intr_arg_t)*SDLA_MAX_PORTS);
		pci_cfg_nxt->irq		= pa->pa_intrline;
	    	pci_cfg_nxt->u.pci.cpu_no	= SDLA_CPU_B;
	    	pci_cfg_nxt->u.pci.slot_no	= pa->pa_device;
	    	pci_cfg_nxt->u.pci.bus_no	= pa->pa_bus;
		for(i=0;i<line_no;i++){
			sc->sc_ih = pci_intr_establish(
					pc, ih, IPL_NET,
					sdla_pci_intr,
					&pci_cfg_nxt->intr_arg[i],
					"sdla");
			if (sc->sc_ih == NULL){
				printf(": couldn't establish interrupt for cpu B");
				if (intrstr != NULL){
					printf(" at %s", intrstr);
				}
				printf("\n");
				return;
			}
			pci_cfg_nxt->intr_arg[i].sd	= (void*)pci_cfg;
			pci_cfg_nxt->intr_arg[i].line_no= i;
			pci_cfg_nxt->intr_arg[i].ready	= 1;
		}
	}
	if (intrstr != NULL){
		printf(" %s\n", intrstr);
	}
	switch(PCI_VENDOR(pa->pa_id)){
	case V3_VENDOR_ID:
		switch(adapter_type){
		case S5141_ADPTR_1_CPU_SERIAL:
    			DEBUG_EVENT("%s: Sangoma S5141/FT1 (Single CPU) adapter\n", 
							self->dv_xname);
			break;
		case S5142_ADPTR_2_CPU_SERIAL:
			DEBUG_EVENT("%s: Sangoma S5142 (Dual CPU) adapter\n", 
							self->dv_xname);
			break;
		case S5143_ADPTR_1_CPU_FT1:
    			DEBUG_EVENT("%s: Sangoma S5143 (Single CPU) FT1 adapter\n", 
							self->dv_xname);
			break;
		case S5144_ADPTR_1_CPU_T1E1:
		case S5148_ADPTR_1_CPU_T1E1:
    			DEBUG_EVENT("%s: Sangoma S5144 (Single CPU) T1/E1 adapter\n", 
							self->dv_xname);
			break;
		case S5145_ADPTR_1_CPU_56K:		
    			DEBUG_EVENT("%s: Sangoma S5145 (Single CPU) 56K adapter\n", 
							self->dv_xname);
			break;
		case S5147_ADPTR_2_CPU_T1E1:
    			DEBUG_EVENT("%s: Sangoma S5147 (Dual CPU) T1/E1 adapter\n", 
							self->dv_xname);
			break;
		}
		break;

	case PCI_VENDOR_ID_GSI:
    		DEBUG_EVENT("%s: Sangoma S518 ADSL adapter\n", 
					self->dv_xname);
		break;

	case SANGOMA_PCI_VENDOR_OLD:
	case SANGOMA_PCI_VENDOR:
		switch(subvendor_id){
		case A101_1TE1_SUBSYS_VENDOR:
    			DEBUG_EVENT("%s: Sangoma AFT-A101 T1/E1 adapter\n", 
							self->dv_xname);
			break;
		case A101_2TE1_SUBSYS_VENDOR:
    			DEBUG_EVENT("%s: Sangoma AFT-A102 T1/E1 adapter\n", 
							self->dv_xname);
			break;
		case A104_4TE1_SUBSYS_VENDOR:
    			DEBUG_EVENT("%s: Sangoma AFT-A104 T1/E1 adapter\n", 
							self->dv_xname);
			break;
		case A104_1TE1_SHARK_SUBSYS_VENDOR:
    			DEBUG_EVENT("%s: Sangoma AFT-A101d T1/E1 adapter\n", 
							self->dv_xname);
			break;
		case A104_2TE1_SHARK_SUBSYS_VENDOR:
    			DEBUG_EVENT("%s: Sangoma AFT-A102d T1/E1 adapter\n", 
							self->dv_xname);
			break;
		case A104_4TE1_SHARK_SUBSYS_VENDOR:
    			DEBUG_EVENT("%s: Sangoma AFT-A104d T1/E1 adapter\n", 
							self->dv_xname);
			break;
		case A104_8TE1_SHARK_SUBSYS_VENDOR:
    			DEBUG_EVENT("%s: Sangoma AFT-A108d T1/E1 adapter\n", 
							self->dv_xname);
			break;
		case A200_REMORA_SHARK_SUBSYS_VENDOR:
    			DEBUG_EVENT("%s: Sangoma AFT-A200 Remora adapter\n", 
							self->dv_xname);
			break;
		}		
		break;
	}	
	return;
}

/*
 ****************************************************************************
 * sdla_pci_intr() - Dummy PCI device interrupt handler.
 */
int sdla_pci_intr(void *arg)
{
#if 0
	sdladev_t	*pci_cfg;
#endif
	sdla_intr_arg_t	*intr_arg = (sdla_intr_arg_t*)arg;

	if (intr_arg == NULL){
		log(LOG_ERR, "%s?: received interrupt from unknown device!\n",
						SDLA_DEV_NAME);
		return 0;
	}
#if 0
	pci_cfg = intr_arg->sd;
	log(LOG_INFO, "sdla_pci: Interrupt from %X Adapter (%d:%p:%p)!\n",
					pci_cfg->subvendor_id,
					intr_arg->line_no,
					intr_arg->isr,
					intr_arg->arg); 
#endif
	if (intr_arg->isr != NULL && intr_arg->arg != NULL){
		intr_arg->isr(intr_arg->arg);
	}
	return 1;
}

/****** End *****************************************************************/
