/*
 * Copyright (c) 2006
 *	Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Alex Feldman.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Alex Feldman AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Alex Feldman OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: sdla_8te1.c,v 1.1 2006/02/24 14:32:41 sangoma Exp $
 */

/*
 ******************************************************************************
 * sdla_8te1.c	WANPIPE(tm) Multiprotocol WAN Link Driver. 
 *				8 ports T1/E1 board configuration.
 *
 * Author: 	Alex Feldman  <al.feldman@sangoma.com>
 *
 * ============================================================================
 * Feb 18, 2006	Alex Feldma	Initial version.
 ******************************************************************************
 */

/******************************************************************************
*			   INCLUDE FILES
******************************************************************************/
#if defined(__NetBSD__) || defined(__FreeBSD__) || defined(__OpenBSD__)
# include <net/wanpipe_includes.h>
# if !defined(CONFIG_PRODUCT_WANPIPE_GENERIC)
#  include <net/wanpipe_snmp.h>
# endif
# include <net/sdla_te1_ds.h>
# include <net/wanpipe.h>	/* WANPIPE common user API definitions */
# include <net/wanproc.h>
#elif (defined __WINDOWS__)
# include <include_A104\pnp\wanpipe_includes.h>
# include <include_A104\pnp\wanpipe.h>	/* WANPIPE common user API definitions */
# include <include_A104\pnp\wanpipe_snmp.h>
# include <include_A104\pnp\sdla_te1_pmc.h>
# if defined te_cfg
#  undef te_cfg
# endif
#elif (defined __LINUX__) || (defined __KERNEL__)
# include <linux/wanpipe_includes.h>
# include <linux/wanpipe_defines.h>
# include <linux/wanpipe_debug.h>
# include <linux/wanproc.h>
# if !defined(CONFIG_PRODUCT_WANPIPE_GENERIC)
#  include <linux/wanpipe_snmp.h>
# endif
# include <linux/sdla_te1_ds.h>
# include <linux/wanpipe.h>	/* WANPIPE common user API definitions */
#else
# error "No OS Defined"
#endif

/******************************************************************************
*			  DEFINES AND MACROS
******************************************************************************/
#define CLEAR_REG(sreg,ereg) {				\
	unsigned short reg;				\
	for(reg = sreg; reg < ereg; reg++){		\
		WRITE_REG(reg, 0x00);			\
	}						\
	}

#define IS_GLREG(reg)	((reg) >= 0xF9 && (reg) <= 0xFF)
#define IS_FRREG(reg)	((reg) <= 0x1F0)
#define IS_LIUREG(reg)	((reg) >= 0x1000 && (reg) <= 0x101F)
#define IS_BERTREG(reg)	((reg) >= 0x1100 && (reg) <= 0x110F)

#define DLS_PORT_DELTA(reg)			\
		IS_GLREG(reg)	? 0x001 :	\
		IS_FRREG(reg)	? 0x200 :	\
		IS_LIUREG(reg)	? 0x020 :	\
		IS_BERTREG(reg)	? 0x010 : 0x001

/* Read/Write to front-end register */
//	DEBUG_EVENT("ADBG: REG:%X<-%X\n", (reg),(val));	
#define WRITE_REG(reg,val)						\
	fe->write_fe_reg(						\
		fe->card,						\
		(int)((reg) + ((fe->fe_cfg.line_no)*(DLS_PORT_DELTA(reg)))),\
		(int)(val))

//DEBUG_EVENT("ADDBG: WRL: %d: %X (%X) <-%X\n",fe_line_no, reg, (int)((int)(reg) + (int)(fe_line_no)*DLS_PORT_DELTA(reg)),val);
#define WRITE_REG_LINE(fe_line_no, reg,val)				\
	fe->write_fe_reg(						\
		fe->card,						\
		(int)((reg) + (fe_line_no)*(DLS_PORT_DELTA(reg))),	\
		(int)(val))
	
#define READ_REG(reg)							\
	fe->read_fe_reg(						\
		fe->card,						\
		(int)((reg) + ((fe->fe_cfg.line_no)*(DLS_PORT_DELTA(reg)))))
	
#define READ_REG_LINE(fe_line_no, reg)					\
	fe->read_fe_reg(						\
		fe->card,						\
		(int)((reg) + (fe_line_no)*(DLS_PORT_DELTA(reg))))

#define IS_T1_ALARM(alarm)			\
		(alarm & 			\
			(			\
			 WAN_TE_BIT_RED_ALARM |	\
			 WAN_TE_BIT_AIS_ALARM |	\
			 WAN_TE_BIT_OOF_ALARM |	\
			 WAN_TE_BIT_LOS_ALARM |	\
			 WAN_TE_BIT_ALOS_ALARM	\
			 )) 

#define IS_E1_ALARM(alarm)			\
		(alarm &	 		\
			(			\
			 WAN_TE_BIT_RED_ALARM | \
			 WAN_TE_BIT_AIS_ALARM | \
			 WAN_TE_BIT_OOF_ALARM | \
			 WAN_TE_BIT_LOS_ALARM | \
			 WAN_TE_BIT_ALOS_ALARM 	\
			 ))


/******************************************************************************
*			  STRUCTURES AND TYPEDEFS
******************************************************************************/


/******************************************************************************
*			  GLOBAL VERIABLES
******************************************************************************/

/******************************************************************************
*			  FUNCTION PROTOTYPES
******************************************************************************/
static int sdla_8te1_global_config(void* pfe);	/* Change to static */
static int sdla_8te1_global_unconfig(void* pfe);	/* Change to static */
static int sdla_8te1_config(void* pfe);	/* Change to static */
static int sdla_8te1_unconfig(void* pfe);	/* Change to static */
static int sdla_8te1_TxChanCtrl(sdla_fe_t* fe, int channel, int enable);
static int sdla_8te1_RxChanCtrl(sdla_fe_t* fe, int channel, int enable);
static int sdla_8te1_intr_ctrl(sdla_fe_t*, int, unsigned char, int); 
static int sdla_8te1_check_intr(sdla_fe_t *fe); 
static int sdla_8te1_intr(sdla_fe_t *fe); 
static int sdla_8te1_udp(sdla_fe_t *fe, void* p_udp_cmd, unsigned char* data);
static int sdla_8te1_flush_pmon(sdla_fe_t *fe);
static int sdla_8te1_pmon(sdla_fe_t *fe);
static int sdla_8te1_polling(sdla_fe_t* fe);
static unsigned long sdla_8te1_read_alarms(sdla_fe_t *fe, int read);
static int sdla_8te1_set_alarms(sdla_fe_t* fe, unsigned long alarms);
static int sdla_8te1_clear_alarms(sdla_fe_t* fe, unsigned long alarms);
static void sdla_8te1_set_status(sdla_fe_t* fe, unsigned long alarms);
static void sdla_8te1_set_intr(sdla_fe_t*);
static void sdla_8te1_clear_intr(sdla_fe_t*);
static int sdla_8te1_print_alarms(sdla_fe_t*, unsigned long);
static int sdla_8te1_set_lbmode(sdla_fe_t*, unsigned char, unsigned char); 
static int sdla_8te1_rbs_update(sdla_fe_t* fe, int, unsigned char);
static int sdla_8te1_set_rbsbits(sdla_fe_t *fe, int, unsigned char);
static int sdla_8te1_rbs_report(sdla_fe_t* fe);
static int sdla_8te1_check_rbsbits(sdla_fe_t* fe, int, unsigned long, int);
static unsigned char sdla_8te1_read_rbsbits(sdla_fe_t* fe, int, int);
static void sdla_8te1_enable_timer(sdla_fe_t*, unsigned char, unsigned long);

#if defined(__NetBSD__) || defined(__FreeBSD__) || defined(__OpenBSD__)
static void sdla_8te1_timer(void* pfe);
#elif defined(__WINDOWS__)
static void sdla_8te1_timer(IN PKDPC Dpc, void* pfe, void* arg2, void* arg3);
#else
static void sdla_8te1_timer(unsigned long pfe);
#endif

static int sdla_8te1_update_alarm_info(sdla_fe_t*, struct seq_file*, int*);
static int sdla_8te1_update_pmon_info(sdla_fe_t*, struct seq_file*, int*);

/******************************************************************************
*			  FUNCTION DEFINITIONS
******************************************************************************/

/******************************************************************************
 *				sdla_te3_get_fe_status()	
 *
 * Description:
 * Arguments:	
 * Returns:
 ******************************************************************************
 */
static char* sdla_8te1_get_fe_media_string(void)
{
	return ("AFT-A108 T1/E1");
}

/******************************************************************************
 *				sdla_8te1_get_fe_status()	
 *
 * Description:
 * Arguments:	
 * Returns:
 ******************************************************************************
 */
static unsigned char sdla_8te1_get_fe_media(sdla_fe_t *fe)
{
	return fe->fe_cfg.media;
}

/******************************************************************************
 *				sdla_8te1_get_fe_status()	
 *
 * Description:
 * Arguments:	
 * Returns:
 ******************************************************************************
 */
static int sdla_8te1_get_fe_status(sdla_fe_t *fe, unsigned char *status)
{
	*status = fe->fe_status;
	return 0;
}

/*
 ******************************************************************************
 *				sdla_8te1_TxChanCtrl()	
 *
 * Description:
 * Arguments:
 * Returns:
 ******************************************************************************
 */
static int sdla_8te1_TxChanCtrl(sdla_fe_t* fe, int channel, int enable)	
{
	int		i = channel / 8, reg;
	int		bit = channel % 8;
	unsigned char	value;

	switch(i){
	case 0:
		reg = REG_TGCCS1;
		break;
	case 1:
		reg = REG_TGCCS2;
		break;
	case 2:
		reg = REG_TGCCS3;
		break;
	case 3:
		reg = REG_TGCCS4;
		break;
	default:
		DEBUG_EVENT("%s: Internal Error (%s:%d)\n",
				fe->name, __FUNCTION__,__LINE__);
		return -EINVAL;

	}
	value = READ_REG(reg);
	if (enable){
		value &= ~(1 << (bit-1));
	}else{
		value |= (1 << (bit-1));
	}
	WRITE_REG(reg, value);
	return 0;
}

/*
 ******************************************************************************
 *				sdla_8te1_RxChanCtrl()	
 *
 * Description:
 * Arguments:
 * Returns:
 ******************************************************************************
 */
static int sdla_8te1_RxChanCtrl(sdla_fe_t* fe, int channel, int enable)
{
	int		i = channel / 8, reg;
	int		bit = channel % 8;
	unsigned char	value;

	switch(i){
	case 0:
		reg = REG_RGCCS1;
		break;
	case 1:
		reg = REG_RGCCS2;
		break;
	case 2:
		reg = REG_RGCCS3;
		break;
	case 3:
		reg = REG_RGCCS4;
		break;
	default:
		DEBUG_EVENT("%s: Internal Error (%s:%d)\n",
				fe->name, __FUNCTION__,__LINE__);
		return -EINVAL;
	}
	value = READ_REG(reg);
	if (enable){
		value &= ~(1 << (bit-1));
	}else{
		value |= (1 << (bit-1));
	}
	WRITE_REG(reg, value);
	return 0;
}


int sdla_8te1_iface_init(void *p_fe_iface)
{
	sdla_fe_iface_t	*fe_iface = (sdla_fe_iface_t*)p_fe_iface;

	fe_iface->global_config		= &sdla_8te1_global_config;
	fe_iface->global_unconfig	= &sdla_8te1_global_unconfig;
	fe_iface->config		= &sdla_8te1_config;
	fe_iface->unconfig		= &sdla_8te1_unconfig;
	fe_iface->isr			= &sdla_8te1_intr;
	fe_iface->check_isr		= &sdla_8te1_check_intr;
	fe_iface->intr_ctrl		= &sdla_8te1_intr_ctrl;
	fe_iface->polling		= &sdla_8te1_polling;
	fe_iface->process_udp		= &sdla_8te1_udp;

	fe_iface->print_fe_alarm	= &sdla_8te1_print_alarms;
	//fe_iface->print_fe_act_channels	= &sdla_te_print_channels;
	fe_iface->read_alarm		= &sdla_8te1_read_alarms;
	//fe_iface->set_fe_alarm		= &sdla_te_set_alarms;
	fe_iface->read_pmon		= &sdla_8te1_pmon;
	fe_iface->flush_pmon		= &sdla_8te1_flush_pmon;
	fe_iface->get_fe_status		= &sdla_8te1_get_fe_status;
	fe_iface->get_fe_media		= &sdla_8te1_get_fe_media;
	fe_iface->get_fe_media_string	= &sdla_8te1_get_fe_media_string;
	fe_iface->update_alarm_info	= &sdla_8te1_update_alarm_info;
	fe_iface->update_pmon_info	= &sdla_8te1_update_pmon_info;
	fe_iface->set_fe_lbmode		= &sdla_8te1_set_lbmode;
	fe_iface->read_rbsbits		= &sdla_8te1_read_rbsbits;
	fe_iface->check_rbsbits		= &sdla_8te1_check_rbsbits;
	fe_iface->report_rbsbits	= &sdla_8te1_rbs_report;
	fe_iface->set_rbsbits		= &sdla_8te1_set_rbsbits;

#if 0
	fe_iface->set_fe_sigctrl	= &sdla_te_sigctrl;
	fe_iface->led_ctrl		= &sdla_te_led_ctrl;
#endif
	return 0;
}

/*
 ******************************************************************************
 *				sdla_8te_global_config()	
 *
 * Description: Global configuration for Sangoma TE1 PMC board.
 * 		Note: 	These register should be program only once for AFT-OCTAL
 * 			cards.
 * Arguments:	
 * Returns:	WANTRUE - TE1 configred successfully, otherwise WAN_FALSE.
 ******************************************************************************
 */
static int sdla_8te1_global_config(void* pfe)
{
	sdla_fe_t	*fe = (sdla_fe_t*)pfe;

	DEBUG_TE1("%s: %s Global Front End configuration\n", 
			fe->name, FE_MEDIA_DECODE(fe));
	WRITE_REG_LINE(0, REG_GTCCR, 0x00);
	WRITE_REG_LINE(0, REG_GTCR1, 0x00);  

	WRITE_REG_LINE(0, REG_GLSRR, 0xFF);  
	WRITE_REG_LINE(0, REG_GFSRR, 0xFF);  
	WP_DELAY(1000);
	WRITE_REG_LINE(0, REG_GLSRR, 0x00);  
	WRITE_REG_LINE(0, REG_GFSRR, 0x00);  

	return 0;
}

/*
 ******************************************************************************
 *				sdla_8te1_global_unconfig()	
 *
 * Description: Global configuration for Sangoma TE1 PMC board.
 * 		Note: 	These register should be program only once for AFT-QUAD
 * 			cards.
 * Arguments:	
 * Returns:	WANTRUE - TE1 configred successfully, otherwise WAN_FALSE.
 ******************************************************************************
 */
static int sdla_8te1_global_unconfig(void* pfe)
{
	sdla_fe_t	*fe = (sdla_fe_t*)pfe;

	DEBUG_EVENT("%s: T1/E1 Global unconfiguration!\n", fe->name);
	WRITE_REG_LINE(0, REG_GLSRR, 0xFF);  
	WRITE_REG_LINE(0, REG_GFSRR, 0xFF);  
	WP_DELAY(1000);
	return 0;
}

/*
 ******************************************************************************
 *				sdla_8te_config()	
 *
 * Description: Configure Sangoma 8 ports T1/E1 board
 * Arguments:	
 * Returns:	WANTRUE - TE1 configred successfully, otherwise WAN_FALSE.
 ******************************************************************************
 */
static int sdla_8te1_config(void* pfe)
{

	sdla_fe_t	*fe = (sdla_fe_t*)pfe;
	unsigned char	value = 0x00;

	WAN_ASSERT(fe->write_fe_reg == NULL);
	WAN_ASSERT(fe->read_fe_reg == NULL);

	/* Initial FE state */
	fe->fe_status=FE_DISCONNECTED;

	/* Revision/Chip ID (Reg. 0x0D) */
	value = READ_REG_LINE(0, REG_IDR);
	fe->fe_chip_id = value & DEVICE_ID_MASK;
	switch(fe->fe_chip_id){
	case DEVICE_ID_DS26528:
		if (WAN_FE_LINENO(fe) < 0 || WAN_FE_LINENO(fe) > 8){
			DEBUG_EVENT(
			"%s: TE Config: Invalid Port selected %d (Min=1 Max=8)\n",
					fe->name,
					WAN_FE_LINENO(fe)+1);
			return -EINVAL;
		}
		break;
	default:
		DEBUG_EVENT("%s: ERROR: Unsupported %s CHIP (0x%02X)\n",
				fe->name, 
				FE_MEDIA_DECODE(fe),
				(fe->fe_chip_id >> 5));
		return -EINVAL;
	}

       	DEBUG_TE1("%s: %s Line decoding %s\n", 
			fe->name, 
			FE_MEDIA_DECODE(fe), 
			FE_LCODE_DECODE(fe));
       	DEBUG_TE1("%s: %s Frame type %s\n", 
			fe->name, 
			FE_MEDIA_DECODE(fe), 
			FE_FRAME_DECODE(fe));
	DEBUG_TE1("%s: %s LBO %s\n", 
			fe->name,
			FE_MEDIA_DECODE(fe), 
			TE_LBO_DECODE(fe));
	if (WAN_TE1_CLK(fe) == WAN_MASTER_CLK){
		if (WAN_TE1_REFCLK(fe) != WAN_TE1_REFCLK_OSC){
       			DEBUG_TE1("%s: %s Clock mode Master (refclk from Port %d)\n", 
					fe->name, 
					FE_MEDIA_DECODE(fe), 
					WAN_TE1_REFCLK(fe)); 
		}else{
       			DEBUG_TE1("%s: %s Clock mode Master (Osciliator)\n", 
					fe->name, 
					FE_MEDIA_DECODE(fe));
		}
	}else{
       		DEBUG_TE1("%s: %s Clock mode Normal\n", 
				fe->name,
				FE_MEDIA_DECODE(fe)); 
	}

	/* Init Rx Framer registers */
	CLEAR_REG(0x0000, 0x00F0);
	/* Init Tx Framer registers */
	CLEAR_REG(0x0100, 0x01F0);
	/* Init LIU registers */
	CLEAR_REG(0x1000, 0x1020);
	/* Init BERT registers */
	CLEAR_REG(0x1100, 0x1110);

	/* Set Rx Framer soft reset */
	WRITE_REG(REG_RMMR, BIT_RMMR_SFTRST);
	/* Set Tx Framer soft reset */
	WRITE_REG(REG_TMMR, BIT_RMMR_SFTRST);
	if (IS_T1_FEMEDIA(fe)){
		/* Clear Rx Framer soft reset */
		WRITE_REG(REG_RMMR, 0x00);
		/* Clear Tx Framer soft reset */
		WRITE_REG(REG_TMMR, 0x00);
		/* Enable Rx Framer */
		WRITE_REG(REG_RMMR, BIT_RMMR_FRM_EN);
		/* Enable Tx Framer */
		WRITE_REG(REG_TMMR, BIT_TMMR_FRM_EN);
	}else{
		/* Clear Rx Framer soft reset */
		WRITE_REG(REG_RMMR, BIT_RMMR_T1E1);
		/* Clear Tx Framer soft reset */
		WRITE_REG(REG_TMMR, BIT_TMMR_T1E1);
		/* Enable Rx Framer */
		WRITE_REG(REG_RMMR, (BIT_RMMR_FRM_EN | BIT_RMMR_T1E1));
		/* Enable Tx Framer */
		WRITE_REG(REG_TMMR, (BIT_TMMR_FRM_EN | BIT_TMMR_T1E1));
	}

	if (IS_T1_FEMEDIA(fe)){
		WRITE_REG(REG_RCR1, BIT_RCR1_T1_SYNCT); 
	}
	switch(WAN_FE_FRAME(fe)){
	case WAN_FR_D4:
		value = READ_REG(REG_RCR1);
		WRITE_REG(REG_RCR1, value | BIT_RCR1_T1_RFM | BIT_RCR1_T1_SYNCC);

		value = READ_REG(REG_TCR3);
		WRITE_REG(REG_TCR3, 
			value | BIT_TCR3_TCSS1 | BIT_TCR3_TCSS0 | BIT_TCR3_TFM);
		break;

	case WAN_FR_ESF:
		value = READ_REG(REG_RCR1);
		value |= BIT_RCR1_T1_SYNCC;
		value &= ~BIT_RCR1_T1_RFM;
		WRITE_REG(REG_RCR1, value);

		value = READ_REG(REG_TCR3);
		value &= ~BIT_TCR3_TFM;
		WRITE_REG(REG_TCR3, value | BIT_TCR3_TCSS1 | BIT_TCR3_TCSS0);
		break;
	case WAN_FR_NCRC4:
		value = READ_REG(REG_TCR3);
		WRITE_REG(REG_TCR3, value | BIT_TCR3_TCSS1 | BIT_TCR3_TCSS0);
		break;
	case WAN_FR_CRC4:
		value = READ_REG(REG_RCR1);
		WRITE_REG(REG_RCR1, value | BIT_RCR1_E1_RCRC4);
		value = READ_REG(REG_TCR1);
		WRITE_REG(REG_TCR1, value | BIT_TCR1_E1_TCRC4);
		value = READ_REG(REG_TCR3);
		WRITE_REG(REG_TCR3, value | BIT_TCR3_TCSS1 | BIT_TCR3_TCSS0);
		break;
	default:
		DEBUG_EVENT("%s: Unsupported Frame mode (%X)\n",
				fe->name, WAN_FE_FRAME(fe));
		return -EINVAL;
	}
	
	if (IS_E1_FEMEDIA(fe)){
		/* CCS signalling mode */
		value = READ_REG(REG_RCR1);
		value |= BIT_RCR1_E1_RSIGM;
		WRITE_REG(REG_RCR1, value);
	}

	switch(WAN_FE_LCODE(fe)){
	case WAN_LCODE_B8ZS:
		value = READ_REG(REG_RCR1);
		WRITE_REG(REG_RCR1, value | BIT_RCR1_T1_RB8ZS);
		value = READ_REG(REG_TCR1);
		WRITE_REG(REG_TCR1, value | BIT_TCR1_T1_TB8ZS);
		break;

	case WAN_LCODE_HDB3:
		value = READ_REG(REG_RCR1);
		WRITE_REG(REG_RCR1, value | BIT_RCR1_E1_RHDB3);
		value = READ_REG(REG_TCR1);
		WRITE_REG(REG_TCR1, value | BIT_TCR1_E1_THDB3);
		break;

	case WAN_LCODE_AMI:
		if (IS_T1_FEMEDIA(fe)){
			value = READ_REG(REG_RCR1);
			WRITE_REG(REG_RCR1, value & ~BIT_RCR1_T1_RB8ZS);
			value = READ_REG(REG_TCR1);
			WRITE_REG(REG_TCR1, value & ~BIT_TCR1_T1_TB8ZS);
		}else{
			value = READ_REG(REG_RCR1);
			WRITE_REG(REG_RCR1, value & ~BIT_RCR1_E1_RHDB3);
			value = READ_REG(REG_TCR1);
			WRITE_REG(REG_TCR1, value & ~BIT_TCR1_E1_THDB3);
		}
		break;

	default:
		DEBUG_EVENT("%s: Unsupported Line code mode (%X)\n",
				fe->name, WAN_FE_LCODE(fe));
		return -EINVAL;
	}

	/* RSYSCLK output */
	WRITE_REG(REG_RIOCR, BIT_RIOCR_RSCLKM);
	/* TSYSCLK input */
	if (IS_T1_FEMEDIA(fe)){
		WRITE_REG(REG_TIOCR, 0x00);
	}else{
		WRITE_REG(REG_TIOCR, BIT_TIOCR_TSCLKM);
	}

	if (IS_E1_FEMEDIA(fe)){
		WRITE_REG(REG_E1TAF, 0x1B);
		WRITE_REG(REG_E1TNAF, 0x40);
	}

	/* Set INIT_DONE */
	value = READ_REG(REG_RMMR);
	WRITE_REG(REG_RMMR, value | BIT_RMMR_INIT_DONE);
	/* Set INIT_DONE */
	value = READ_REG(REG_TMMR);
	WRITE_REG(REG_TMMR, value | BIT_TMMR_INIT_DONE);

	/* T1/J1 or E1 */
	if (IS_T1_FEMEDIA(fe)){
		WRITE_REG(REG_LTRCR, BIT_LTRCR_T1J1E1S);
	}else{
		/* E1 | G.775 LOS */
		WRITE_REG(REG_LTRCR, 0x00);
	}

	value = 0x00;
	switch(WAN_TE1_LBO(fe)) {
	case WAN_T1_LBO_0_DB:
		value = 0x00;
		break;
	case WAN_T1_LBO_75_DB:
		value = BIT_LTITSR_L2 | BIT_LTITSR_L0;
		break;
	case WAN_T1_LBO_15_DB:
		value = BIT_LTITSR_L2 | BIT_LTITSR_L1;
		break;
	case WAN_T1_LBO_225_DB:
		value = BIT_LTITSR_L2 | BIT_LTITSR_L1 | BIT_LTITSR_L0;
		break;
	case WAN_T1_0_133:
		value = 0x00;
		break;
	case WAN_T1_133_266:
		value = BIT_LTITSR_L0;
		break;
	case WAN_T1_266_399:
		value = BIT_LTITSR_L1;
		break;
	case WAN_T1_399_533:
		value = BIT_LTITSR_L1 | BIT_LTITSR_L0;
		break;
	case WAN_T1_533_655:
		value = BIT_LTITSR_L2;
		break;
	case WAN_E1_120:
		value = BIT_LTITSR_L0;
		break;
	case WAN_E1_75:
		value = 0x00;
		break;
	}
	if (IS_T1_FEMEDIA(fe)){
		WRITE_REG(REG_LTITSR, value | BIT_LTITSR_TIMPL0);
	}else{
		WRITE_REG(REG_LTITSR,
			value | BIT_LTITSR_TIMPL1 | BIT_LTITSR_TIMPL0);
	}

	value = BIT_LRISMR_RSMS1 | BIT_LRISMR_RSMS0;
	if (IS_T1_FEMEDIA(fe)){
		WRITE_REG(REG_LRISMR, value | BIT_LRISMR_RIMPM0);
	}else{
		if (WAN_TE1_LBO(fe) == WAN_E1_120){
			value |= BIT_LRISMR_RIMPM1 | BIT_LRISMR_RIMPM0;
		}
		WRITE_REG(REG_LRISMR, value);
	}
		
	/* extra configuration here */
#if 1
	if (WAN_TE1_ACTIVE_CH(fe) != ENABLE_ALL_CHANNELS){
		int channel_range = (IS_T1_FEMEDIA(fe)) ? 
				NUM_OF_T1_CHANNELS : NUM_OF_E1_TIMESLOTS;
		int i = 0;

 		for(i = 1; i <= channel_range; i++){
			if (WAN_TE1_ACTIVE_CH(fe) & (1 << (i - 1))){
				DEBUG_EVENT("%s: Disable channel %d\n", 
						fe->name, i);
				sdla_8te1_TxChanCtrl(fe, i, 0);
				sdla_8te1_RxChanCtrl(fe, i, 0);
			}
		}
	}
#endif
	/* Turn on LIU output */
	WRITE_REG(REG_LMCR, BIT_LMCR_TE);

	/* Initialize and start T1/E1 timer */
	wan_set_bit(TE_TIMER_KILL,(void*)&fe->fe_param.te_param.critical);
	wan_init_timer(
		&fe->fe_param.te_param.timer, 
		sdla_8te1_timer,
	       	(wan_timer_arg_t)fe);

	wan_set_bit(TE_CONFIGURED,(void*)&fe->fe_param.te_param.critical);

	/* Enable Basic Interrupt */
	sdla_8te1_set_intr(fe);
			
	/* Initialize T1/E1 timer */
	wan_clear_bit(TE_TIMER_KILL,(void*)&fe->fe_param.te_param.critical);

	/* Start T1/E1 timer */
	sdla_8te1_enable_timer(fe, TE_LINKDOWN_TIMER, POLLING_TE1_TIMER);

	return 0;

}

/*
 ******************************************************************************
 *			sdla_8te1_unconfig()	
 *
 * Description: T1/E1 unconfig.
 * Arguments:
 * Returns:
 ******************************************************************************
 */
static int sdla_8te1_unconfig(void* pfe)
{
	sdla_fe_t	*fe = (sdla_fe_t*)pfe;

	if (!wan_test_bit(TE_CONFIGURED,(void*)&fe->fe_param.te_param.critical)){
		return -EINVAL;
	}

	/* Clear and Kill TE timer poll command */
	wan_clear_bit(TE_CONFIGURED,(void*)&fe->fe_param.te_param.critical);
	wan_set_bit(TE_TIMER_KILL,(void*)&fe->fe_param.te_param.critical);

	/* FIXME: Alex to disable interrupts here */
	sdla_8te1_clear_intr(fe);

	wan_del_timer(&fe->fe_param.te_param.timer);

	fe->fe_param.te_param.timer_cmd = 0x00;
	wan_clear_bit(TE_TIMER_RUNNING,(void*)&fe->fe_param.te_param.critical);


	return 0;
}

/*
 ******************************************************************************
 *			sdla_8te1_set_intr()	
 *
 * Description: Enable T1/E1 interrupts.
 * Arguments:
 * Returns:
 ******************************************************************************
 */
static void sdla_8te1_set_intr(sdla_fe_t* fe)
{
	unsigned char	mask;
	DEBUG_EVENT("%s: Enabling TE1 Interrupts \n",fe->name);

	mask = READ_REG(REG_GFIMR);
	WRITE_REG(REG_GFIMR, mask | (1 << WAN_FE_LINENO(fe)));

	WRITE_REG(REG_RIM1,
			BIT_RIM1_RRAIC | BIT_RIM1_RRAID |
			BIT_RIM1_RAISC | BIT_RIM1_RAISD |
			BIT_RIM1_RLOSC | BIT_RIM1_RLOSD |
			BIT_RIM1_RLOFC | BIT_RIM1_RLOFD);
	WRITE_REG(REG_RIM4, BIT_RIM4_RSCOS);

	mask = READ_REG(REG_GLIMR);
	WRITE_REG(REG_GLIMR, mask | (1 << WAN_FE_LINENO(fe)));
	WRITE_REG(REG_LSIMR,
			BIT_LSIMR_OCCIM | BIT_LSIMR_OCDIM |
			BIT_LSIMR_SCCIM | BIT_LSIMR_SCCIM |
			BIT_LSIMR_LOSCIM | BIT_LSIMR_LOSDIM |
			BIT_LSIMR_JALTCIM | BIT_LSIMR_JALTSIM);

	return;
}

/*
 ******************************************************************************
 *			sdla_8te1_clear_intr()	
 *
 * Description: Disable T1/E1 interrupts.
 * Arguments:
 * Returns:
 ******************************************************************************
 */
static void sdla_8te1_clear_intr(sdla_fe_t* fe)
{

	DEBUG_EVENT("%s: Clearing TE1 Interrupts\n",fe->name);

	WRITE_REG(REG_RIM1, 0x00);
	WRITE_REG(REG_RIM2, 0x00);
	WRITE_REG(REG_RIM3, 0x00);
	WRITE_REG(REG_RIM4, 0x00);
	WRITE_REG(REG_RIM5, 0x00);
	WRITE_REG(REG_RIM7, 0x00);

	WRITE_REG(REG_TIM1, 0x00);
	WRITE_REG(REG_TIM2, 0x00);
	WRITE_REG(REG_TIM3, 0x00);

	WRITE_REG(REG_LSIMR, 0x00);
	return;
}


/*
 ******************************************************************************
 *			sdla_8te1_set_status()	
 *
 * Description: Set T1/E1 status. Enable OOF and LCV interrupt (if status 
 * 		changed to disconnected.
 * Arguments:
 * Returns:
 ******************************************************************************
 */
static void sdla_8te1_set_status(sdla_fe_t* fe, unsigned long alarms)
{
	sdla_t		*card = (sdla_t*)fe->card;
	unsigned char	curr_fe_status = fe->fe_status;


	if (fe->fe_alarm & WAN_TE_BIT_OOF_ALARM && fe->fe_alarm & WAN_TE_BIT_LOS_ALARM){
		fe->fe_alarm |= WAN_TE_BIT_RED_ALARM;
	}else{
		fe->fe_alarm &= ~WAN_TE_BIT_RED_ALARM;
	}
	if (IS_T1_FEMEDIA(fe)){
		if (IS_T1_ALARM(alarms)){
			if (fe->fe_status != FE_DISCONNECTED){
				sdla_8te1_set_alarms(fe, WAN_TE_BIT_YEL_ALARM);
				fe->fe_status = FE_DISCONNECTED;
			}
		}else{
			sdla_8te1_clear_alarms(fe, WAN_TE_BIT_YEL_ALARM);
			if (!(fe->fe_alarm & WAN_TE_BIT_YEL_ALARM)){
				if (fe->fe_status != FE_CONNECTED){
					fe->fe_status = FE_CONNECTED;
				}
			}else{
				fe->fe_status = FE_DISCONNECTED;
			}
		}
	}else{
		if (IS_E1_ALARM(alarms)){
			if (fe->fe_status != FE_DISCONNECTED){
				fe->fe_status = FE_DISCONNECTED;
			}
		}else{
			if (fe->fe_status != FE_CONNECTED){
				fe->fe_status = FE_CONNECTED;
			}
		}
	}
	if (curr_fe_status != fe->fe_status){
		if (fe->fe_status == FE_CONNECTED){
			DEBUG_EVENT("%s: %s connected!\n", 
					fe->name,
					FE_MEDIA_DECODE(fe));
		}else{
			DEBUG_EVENT("%s: %s disconnected!\n", 
					fe->name,
					FE_MEDIA_DECODE(fe));
		}

	}

	if (card->wandev.te_report_alarms){
		card->wandev.te_report_alarms(card, alarms);
	}

	return;
}


/*
*******************************************************************************
**				sdla_te_alaram_print()	
**
** Description: 
** Arguments:
** Returns:
*/
static int sdla_8te1_print_alarms(sdla_fe_t* fe, unsigned long alarms)
{
	if (!alarms){
		alarms = fe->fe_alarm; 
	}

	DEBUG_EVENT("%s: %s Alarms:\n",
			fe->name,
			FE_MEDIA_DECODE(fe));
	DEBUG_EVENT("%s:    ALOS: %s\n", 
			fe->name, WAN_TE_ALOS_ALARM(alarms));
	DEBUG_EVENT("%s:    LOS*:  %s\n", 
			fe->name, WAN_TE_LOS_ALARM(alarms));
	DEBUG_EVENT("%s:    RED:  %s\n", 
			fe->name, WAN_TE_RED_ALARM(alarms));
	DEBUG_EVENT("%s:    AIS*:  %s\n", 
			fe->name, WAN_TE_AIS_ALARM(alarms));
	DEBUG_EVENT("%s:    RAI*:  %s\n", 
				fe->name, WAN_TE_RAI_ALARM(alarms));
	DEBUG_EVENT("%s:    OOF*:  %s\n", 
			fe->name, WAN_TE_OOF_ALARM(alarms));
	return 0;
}

static unsigned long sdla_8te1_read_alarms(sdla_fe_t *fe, int read)
{

	if (read){
		unsigned char	status;
		status = READ_REG(REG_RRTS1);
		DEBUG_EVENT("ADBG: %s: Reading alarms %X\n",
				fe->name, status);
		if (status & BIT_RRTS1_RRAI){
			DEBUG_EVENT("%s: RAI is ON\n", fe->name);
			fe->fe_alarm |= WAN_TE_BIT_RAI_ALARM;		
		}else{
			DEBUG_EVENT("%s: RAI is OFF\n", fe->name);
			fe->fe_alarm &= ~WAN_TE_BIT_RAI_ALARM;		
		}
		if (status & BIT_RRTS1_RAIS){
			DEBUG_EVENT("%s: AIS is ON\n", fe->name);
			fe->fe_alarm |= WAN_TE_BIT_AIS_ALARM;		
		}else{
			DEBUG_EVENT("%s: AIS is OFF\n", fe->name);
			fe->fe_alarm &= ~WAN_TE_BIT_AIS_ALARM;		
		}
		if (status & BIT_RRTS1_RLOS){
			DEBUG_EVENT("%s: LOS is ON\n", fe->name);
			fe->fe_alarm |= WAN_TE_BIT_LOS_ALARM;		
		}else{
			DEBUG_EVENT("%s: LOS is OFF\n", fe->name);
			fe->fe_alarm &= ~WAN_TE_BIT_LOS_ALARM;		
		}
		if (status & BIT_RRTS1_RLOF){
			DEBUG_EVENT("%s: LOF is ON\n", fe->name);
			fe->fe_alarm |= WAN_TE_BIT_OOF_ALARM;		
		}else{
			DEBUG_EVENT("%s: LOF is OFF\n", fe->name);
			fe->fe_alarm &= ~WAN_TE_BIT_OOF_ALARM;		
		}
		if (fe->fe_alarm & WAN_TE_BIT_OOF_ALARM && fe->fe_alarm & WAN_TE_BIT_LOS_ALARM){
			DEBUG_EVENT("%s: RED is ON\n", fe->name);
			fe->fe_alarm |= WAN_TE_BIT_RED_ALARM;
		}else{
			DEBUG_EVENT("%s: RED is OFF\n", fe->name);
			fe->fe_alarm &= ~WAN_TE_BIT_RED_ALARM;
		}
	}
	return fe->fe_alarm;
}

/*
*******************************************************************************
**				sdla_8te1_set_alarms()	
**
** Description: 
** Arguments:
** Returns:
*/
static int sdla_8te1_set_alarms(sdla_fe_t* fe, unsigned long alarms)
{
	if (alarms & WAN_TE_BIT_YEL_ALARM){
		DEBUG_EVENT("%s: Setting YELLOW alarm (not supported)!\n",
							fe->name);
	}
	return 0;
}

/*
*******************************************************************************
**				sdla_8te1_clear_alarms()	
**
** Description: 
** Arguments:
** Returns:
*/
static int sdla_8te1_clear_alarms(sdla_fe_t* fe, unsigned long alarms)
{
	if (alarms & WAN_TE_BIT_YEL_ALARM){
		DEBUG_EVENT("%s: Clearing YELLOW alarm (not supported)!\n",
						fe->name);
	}
	return 0;
}

/*
 ******************************************************************************
 *				sdla_8te1_rbs_report()	
 *
 * Description:
 * Arguments:
 * Returns:
 ******************************************************************************
 */
static int sdla_8te1_rbs_report(sdla_fe_t* fe)
{
	DEBUG_EVENT("%s: Internal Error (%s:%d)!\n",
			fe->name, __FUNCTION__,__LINE__);
	return 0;
}


/*
 ******************************************************************************
 *				sdla_8te1_read_rbsbits()	
 *
 * Description:
 * Arguments:	channo: 1-24 for T1
 *			1-32 for E1
 * Returns:
 ******************************************************************************
 */
static unsigned char sdla_8te1_read_rbsbits(sdla_fe_t* fe, int channo, int report)
{
	sdla_t*		card = (sdla_t*)fe->card;
	int		rs_offset = 0, range = 0;
	unsigned char	rbsbits = 0x00, status = 0x00;
	
	if (IS_E1_FEMEDIA(fe)){
		rs_offset = (channo - 1) % 16;
		range = 16;
	}else{
		rs_offset = (channo - 1) % 12;
		range = 12;
	}
	rbsbits = READ_REG(REG_RS1 + rs_offset - 1);
	if (channo <= range){
		rbsbits = (rbsbits >> 4) & 0x0F;
	}else{
		rbsbits &= 0xF;
	}
	
	if (rbsbits & BIT_RS_A) status |= WAN_RBS_SIG_A;
	if (rbsbits & BIT_RS_B) status |= WAN_RBS_SIG_B;
	if (rbsbits & BIT_RS_C) status |= WAN_RBS_SIG_C;
	if (rbsbits & BIT_RS_D) status |= WAN_RBS_SIG_D;

	sdla_8te1_rbs_update(fe, channo, status);

	if (report && card->wandev.te_report_rbsbits){
		card->wandev.te_report_rbsbits(
					card, 
					channo, 
					status);
	}
	return status;
}

/*
 ******************************************************************************
 *				sdla_te_rbs_update()	
 *
 * Description:
 * Arguments:
 * Returns:
 ******************************************************************************
 */
static int 
sdla_8te1_rbs_update(sdla_fe_t* fe, int channo, unsigned char status)
{

	if (fe->fe_debug & WAN_FE_DEBUG_RBS_RX_ENABLE){
		DEBUG_EVENT(
		"%s: %s:%-3d RX RBS A:%1d B:%1d C:%1d D:%1d\n",
					fe->name,
					FE_MEDIA_DECODE(fe),
					channo, 
					(status & WAN_RBS_SIG_A) ? 1 : 0,
					(status & WAN_RBS_SIG_B) ? 1 : 0,
					(status & WAN_RBS_SIG_C) ? 1 : 0,
					(status & WAN_RBS_SIG_D) ? 1 : 0);
	}
	
	/* Update rbs value in private structures */
	wan_set_bit(channo, &fe->fe_param.te_param.rx_rbs_status);
	fe->fe_param.te_param.rx_rbs[channo] = status;

	if (status & WAN_RBS_SIG_A){
		wan_set_bit(channo,
			(unsigned long*)&fe->fe_param.te_param.rx_rbs_A);
	}else{
		wan_clear_bit(channo,
			(unsigned long*)&fe->fe_param.te_param.rx_rbs_A);
	}	
	if (status & WAN_RBS_SIG_B){
		wan_set_bit(channo,
			(unsigned long*)&fe->fe_param.te_param.rx_rbs_B);
	}else{
		wan_clear_bit(channo,
			(unsigned long*)&fe->fe_param.te_param.rx_rbs_B);
	}
	if (status & WAN_RBS_SIG_C){
		wan_set_bit(channo,
			(unsigned long*)&fe->fe_param.te_param.rx_rbs_C);
	}else{
		wan_clear_bit(channo,
			(unsigned long*)&fe->fe_param.te_param.rx_rbs_C);
	}
	if (status & WAN_RBS_SIG_D){
		wan_set_bit(channo,
			(unsigned long*)&fe->fe_param.te_param.rx_rbs_D);
	}else{
		wan_clear_bit(channo,
			(unsigned long*)&fe->fe_param.te_param.rx_rbs_D);
	}

	return 0;
}

/*
 ******************************************************************************
 *				sdla_8te1_check_rbsbits()	
 *
 * Description:
 * Arguments:
 * Returns:
 ******************************************************************************
 */
static int 
sdla_8te1_check_rbsbits(sdla_fe_t* fe, int ch_base, unsigned long ts_map, int report)
{
	sdla_t*		card = (sdla_t*)fe->card;
	unsigned char	rs_status = 0x0, rbsbits = 0x00, bit_mask = 0x00;
	int		rs_reg = 0, rs_offset = 0;
	int		i = 0, channel, range = 12;

	switch(ch_base){
	case 1:
		rs_reg = REG_RSS1;
		break;
	case 9:
		rs_reg = REG_RSS2;
		break;
	case 17:
		rs_reg = REG_RSS3;
		break;
	case 25:
		rs_reg = REG_RSS4;
		break;
	}
	rs_status = READ_REG(rs_reg);
	
	if (rs_status == 0x00){
		return 0;
	}

	if (IS_E1_FEMEDIA(fe)){
		range = 16;
	}
	for(i = 0; i < 8; i ++) {
		channel = ch_base + i;
		if (IS_E1_FEMEDIA(fe)){
			if (channel == 1 || channel == 17){
				continue;
			}
			rs_offset = channel % 16;
		}else{
			rs_offset = (channel - 1) % 12;
		}
		/* If this channel/timeslot is not selected, move to 
		 * another channel/timeslot */
		if (!wan_test_bit(channel-1, &ts_map)){
			continue;
		}
		bit_mask = (1 << i);
		if(rs_status & bit_mask) {
			unsigned char	abcd_status = 0x00;

			rbsbits = READ_REG(REG_RSS1 + rs_offset - 1);
			if (channel > range){
				rbsbits &= 0x0F;
			}else{
				rbsbits = (rbsbits >> 4) & 0x0F;
			}
		
			if (rbsbits & BIT_RS_A) abcd_status |= WAN_RBS_SIG_A;
			if (rbsbits & BIT_RS_B) abcd_status |= WAN_RBS_SIG_B;
			if (rbsbits & BIT_RS_C) abcd_status |= WAN_RBS_SIG_C;
			if (rbsbits & BIT_RS_D) abcd_status |= WAN_RBS_SIG_D;
			if (IS_E1_FEMEDIA(fe)){
				channel--;
			}
			sdla_8te1_rbs_update(fe, channel, abcd_status);
			if (report && card->wandev.te_report_rbsbits){
				card->wandev.te_report_rbsbits(
							card, 
							channel, 
							abcd_status);
			}
			WRITE_REG(rs_reg, bit_mask);
		}
	}	
	return 0; 
}

/*
 ******************************************************************************
 *				sdla_8te1_set_RBS()	
 *
 * Description:
 * Arguments:
 * Returns:
 ******************************************************************************
 */
static int 
sdla_8te1_set_rbsbits(sdla_fe_t *fe, int channel, unsigned char status)
{
	int		rs_offset = 0, range = 0;
	unsigned char	rbsbits = 0x00;

	if (channel > fe->fe_param.te_param.max_channels){
		DEBUG_EVENT("%s: Invalid channel number %d\n",
				fe->name, channel);
		return -EINVAL;
	}

	if (IS_E1_FEMEDIA(fe) && (channel == 0 || channel == 16)){
		DEBUG_EVENT("%s: Invalid channel number %d for Rx SIG\n",
				fe->name, channel);
		return -EINVAL;
	}
	if (status & WAN_RBS_SIG_A) rbsbits |= BIT_TS_A;
	if (status & WAN_RBS_SIG_B) rbsbits |= BIT_TS_B;
	if (status & WAN_RBS_SIG_C) rbsbits |= BIT_TS_C;
	if (status & WAN_RBS_SIG_D) rbsbits |= BIT_TS_D;
	if (fe->fe_debug & WAN_FE_DEBUG_RBS_TX_ENABLE){
		DEBUG_EVENT("%s: %s:%-3d TX RBS A:%1d B:%1d C:%1d D:%1d\n",
				fe->name,
				FE_MEDIA_DECODE(fe),
				channel,	
				(rbsbits & BIT_TS_A) ? 1 : 0,
				(rbsbits & BIT_TS_B) ? 1 : 0,
				(rbsbits & BIT_TS_C) ? 1 : 0,
				(rbsbits & BIT_TS_D) ? 1 : 0);
	}
	if (rbsbits & BIT_TS_A){
		wan_set_bit(channel,(unsigned long*)&fe->fe_param.te_param.tx_rbs_A);
	}else{
		wan_clear_bit(channel,(unsigned long*)&fe->fe_param.te_param.tx_rbs_A);
	}
	if (rbsbits & BIT_TS_B){
		wan_set_bit(channel,(unsigned long*)&fe->fe_param.te_param.tx_rbs_B);
	}else{
		wan_clear_bit(channel,(unsigned long*)&fe->fe_param.te_param.tx_rbs_B);
	}
	if (rbsbits & BIT_TS_C){
		wan_set_bit(channel,(unsigned long*)&fe->fe_param.te_param.tx_rbs_C);
	}else{
		wan_clear_bit(channel,(unsigned long*)&fe->fe_param.te_param.tx_rbs_C);
	}
	if (rbsbits & BIT_TS_D){
		wan_set_bit(channel,(unsigned long*)&fe->fe_param.te_param.tx_rbs_D);
	}else{
		wan_clear_bit(channel,(unsigned long*)&fe->fe_param.te_param.tx_rbs_D);
	}
	if (IS_E1_FEMEDIA(fe)){
		rs_offset = channel % 16;
		range = 15;
	}else{
		rs_offset = (channel - 1) % 12;
		range = 12;
	}
	if (channel <= range){
		rbsbits = rbsbits << 4;
	}
	WRITE_REG(REG_TS1 + rs_offset - 1, rbsbits);
	return 0;
}

/*
 ******************************************************************************
 *				sdla_te_intr()	
 *
 * Description: Check interrupt type.
 * Arguments: 	card 		- pointer to device structure.
 * 		write_register 	- write register function.
 * 		read_register	- read register function.
 * Returns:	None.
 ******************************************************************************
 */
static int
sdla_8te1_intr_ctrl(sdla_fe_t *fe, int dummy, unsigned char type, int enable) 
{

	WAN_ASSERT(fe->write_fe_reg == NULL);
	WAN_ASSERT(fe->read_fe_reg == NULL);
	if (!wan_test_bit(TE_CONFIGURED,(void*)&fe->fe_param.te_param.critical)){
		return 0;
	}
	switch(type){
	case WAN_TE_INTR_SIGNALLING:
		if (enable){
			/* */
			WRITE_REG(REG_RSCSE1, BITS_RSCSE1_ALL);
			WRITE_REG(REG_RSCSE2, BITS_RSCSE2_ALL);
			WRITE_REG(REG_RSCSE3, BITS_RSCSE3_ALL);
			if (IS_E1_FEMEDIA(fe)){
				WRITE_REG(REG_RSCSE4, BITS_RSCSE4_ALL);
			}
			WRITE_REG(REG_SSIE1,  BITS_SSIE1_ALL);
			WRITE_REG(REG_SSIE2,  BITS_SSIE1_ALL);
			WRITE_REG(REG_SSIE3,  BITS_SSIE3_ALL);
			if (IS_E1_FEMEDIA(fe)){
				WRITE_REG(REG_SSIE4,  BITS_SSIE4_ALL);
			}
		}else{
			WRITE_REG(REG_RSCSE1, 0x00);
			WRITE_REG(REG_RSCSE2, 0x00);
			WRITE_REG(REG_RSCSE3, 0x00);
			if (IS_E1_FEMEDIA(fe)){
				WRITE_REG(REG_RSCSE4, 0x00);
			}
			WRITE_REG(REG_SSIE1,  0x00);
			WRITE_REG(REG_SSIE2,  0x00);
			WRITE_REG(REG_SSIE3,  0x00);
			if (IS_E1_FEMEDIA(fe)){
				WRITE_REG(REG_SSIE4,  0x00);
			}
		}
		break;
	default:
		DEBUG_EVENT("%s: Unsupported interrupt type %X!\n",
					fe->name, type);
		return -EINVAL;
	}

	return 0;
}

static int sdla_8te1_check_intr(sdla_fe_t *fe) 
{
	unsigned char	framer_istatus;
	unsigned char	liu_istatus;
	unsigned char	bert_istatus;
	
	WAN_ASSERT(fe->write_fe_reg == NULL);
	WAN_ASSERT(fe->read_fe_reg == NULL);

	if (!wan_test_bit(TE_CONFIGURED,(void*)&fe->fe_param.te_param.critical)){
		return 0;
	}
	
	framer_istatus = READ_REG(REG_GFISR);
	liu_istatus = READ_REG(REG_GLISR);
	bert_istatus = READ_REG(REG_GBISR);

	if (framer_istatus & (1 << WAN_FE_LINENO(fe))){
		DEBUG_ISR("%s: Interrupt for line %d (framer)\n",
					fe->name, WAN_FE_LINENO(fe));
		return 1;
	}
	if (liu_istatus & (1 << WAN_FE_LINENO(fe))){
		DEBUG_ISR("%s: Interrupt for line %d (liu)\n",
					fe->name, WAN_FE_LINENO(fe));
		return 1;
	}
	if (bert_istatus & (1 << WAN_FE_LINENO(fe))){
		DEBUG_ISR("%s: Interrupt for line %d (bert)\n",
					fe->name, WAN_FE_LINENO(fe));
		return 1;
	}
	DEBUG_ISR("%s: This interrupt not for this port %d\n",
				fe->name,
				WAN_FE_LINENO(fe)+1);
	return 0;
}

static int sdla_8te1_framer_rx_intr(sdla_fe_t *fe) 
{
	unsigned char	istatus;
	unsigned char	status;

	istatus = READ_REG(REG_RIIR);
	if (istatus & BIT_RIIR_RLS1){
		status = READ_REG(REG_RLS1);
		if (status & BIT_RLS1_RRAIC){
			fe->fe_alarm &= ~WAN_TE_BIT_RAI_ALARM;		
		}
		if (status & BIT_RLS1_RRAID){
			fe->fe_alarm |= WAN_TE_BIT_RAI_ALARM;		
		}
		if (status & BIT_RLS1_RAISC){
			fe->fe_alarm &= ~WAN_TE_BIT_AIS_ALARM;		
		}
		if (status & BIT_RLS1_RAISD){
			fe->fe_alarm |= WAN_TE_BIT_AIS_ALARM;		
		}
		if (status & BIT_RLS1_RLOSC){
			fe->fe_alarm &= ~WAN_TE_BIT_LOS_ALARM;		
		}
		if (status & BIT_RLS1_RLOSD){
			fe->fe_alarm |= WAN_TE_BIT_LOS_ALARM;		
		}
		if (status & BIT_RLS1_RLOFC){
			fe->fe_alarm &= ~WAN_TE_BIT_OOF_ALARM;		
		}
		if (status & BIT_RLS1_RLOFD){
			fe->fe_alarm |= WAN_TE_BIT_OOF_ALARM;		
		}
		WRITE_REG(REG_RLS1, 0xFF);
	}
	if (istatus & BIT_RIIR_RLS2){
		if (IS_E1_FEMEDIA(fe)){
			status = READ_REG(REG_RLS2);
			if (status & BIT_RLS2_E1_RSA1){
				DEBUG_EVENT(
				"%s: Receive Signalling All Ones Event!\n",
						fe->name);
			}
			if (status & BIT_RLS2_E1_RSA0){
				DEBUG_EVENT(
				"%s: Receive Signalling All Ones Event!\n",
						fe->name);
			}
			if (status & BIT_RLS2_E1_RCMF){
				DEBUG_EVENT(
				"%s: Receive CRC4 Multiframe Event!\n",
						fe->name);
			}
			if (status & BIT_RLS2_E1_RAF){
				DEBUG_EVENT(
				"%s: Receive Align Frame Event!\n",
						fe->name);
			}
		}
		WRITE_REG(REG_RLS2, 0xFF);
	}
	if (istatus & BIT_RIIR_RLS3){
		status = READ_REG(REG_RLS3);
		if (IS_T1_FEMEDIA(fe)){
			if (status & BIT_RLS3_T1_LORCC){
				DEBUG_EVENT(
				"%s: Loss of Receive Clock Condition Clear!\n",
						fe->name);
			}
			if (status & BIT_RLS3_T1_LORCD){
				DEBUG_EVENT(
				"%s: Loss of Receive Clock Condition Detect!\n",
						fe->name);
			}
		}else{
			if (status & BIT_RLS3_E1_LORCC){
				DEBUG_EVENT(
				"%s: Loss of Receive Clock Condition Clear!\n",
						fe->name);
			}
			if (status & BIT_RLS3_E1_LORCD){
				DEBUG_EVENT(
				"%s: Loss of Receive Clock Condition Detect!\n",
						fe->name);
			}
		}
		WRITE_REG(REG_RLS3, 0xFF);
	}
	if (istatus & BIT_RIIR_RLS4){
		status = READ_REG(REG_RLS4);
		if (status & BIT_RLS4_RSCOS){
			DEBUG_EVENT(
			"%s: Receive Signalling status changed!\n",
					fe->name);
			sdla_8te1_check_rbsbits(fe, 1, ENABLE_ALL_CHANNELS, 1); 
			sdla_8te1_check_rbsbits(fe, 9, ENABLE_ALL_CHANNELS, 1);
			sdla_8te1_check_rbsbits(fe, 17, ENABLE_ALL_CHANNELS, 1);
			if (IS_E1_FEMEDIA(fe)){
				sdla_8te1_check_rbsbits(fe, 25, ENABLE_ALL_CHANNELS, 1);
			}
		}
		if (status & BIT_RLS4_TIMER){
			DEBUG_EVENT(
			"%s: Performance monitor counters have been updated!\n",
					fe->name);
			sdla_8te1_pmon(fe);
		}
		WRITE_REG(REG_RLS4, 0xFF);
	}
	if (istatus & BIT_RIIR_RLS5){
		DEBUG_EVENT("%s: Internal Error (%s:%d)!\n",
				fe->name, __FUNCTION__,__LINE__);
		WRITE_REG(REG_RLS5, 0xFF);
	}
	if (istatus & BIT_RIIR_RLS6){
		DEBUG_EVENT("%s: Internal Error (%s:%d)!\n",
				fe->name, __FUNCTION__,__LINE__);
		WRITE_REG(REG_RLS6, 0xFF);
	}
	if (istatus & BIT_RIIR_RLS7){
		DEBUG_EVENT("%s: Internal Error (%s:%d)!\n",
				fe->name, __FUNCTION__,__LINE__);
		WRITE_REG(REG_RLS7, 0xFF);
	}
	
	return 0;
}

static int sdla_8te1_framer_tx_intr(sdla_fe_t *fe) 
{
	unsigned char	istatus;
	unsigned char	status;

	istatus = READ_REG(REG_TIIR);
	if (istatus & BIT_TIIR_TLS1){
		status = READ_REG(REG_TLS1);
		if (status & BIT_TLS1_TPDV){
			DEBUG_EVENT(
			"%s: Transmit Pulse Density Violation Event!\n",
					fe->name);
		}
		if (status & BIT_TLS1_LOTCC){
			DEBUG_EVENT(
			"%s: Loss of Transmit Clock condition Clear!\n",
					fe->name);
		}
		WRITE_REG(REG_TLS1, 0xFF);
	}
	if (istatus & BIT_TIIR_TLS2){
		DEBUG_EVENT("%s: Internal Error (%s:%d)!\n",
				fe->name, __FUNCTION__,__LINE__);
		WRITE_REG(REG_TLS2, 0xFF);
	}
	if (istatus & BIT_TIIR_TLS3){
		DEBUG_EVENT("%s: Internal Error (%s:%d)!\n",
				fe->name, __FUNCTION__,__LINE__);
		WRITE_REG(REG_TLS3, 0xFF);
	}
	return 0;
}

static int sdla_8te1_bert_intr(sdla_fe_t *fe) 
{
	DEBUG_EVENT("%s: Internal Error (%s:%d)!\n",
				fe->name, __FUNCTION__,__LINE__);
	return 0;
}

static int sdla_8te1_liu_intr(sdla_fe_t *fe) 
{
	unsigned char	istatus;

	istatus = READ_REG(REG_LLSR);
	if (istatus & BIT_LLSR_JALTC){
		DEBUG_EVENT("%s: Jitter Attenuator Limit Trip Clear!\n",
					fe->name);
	}
	if (istatus & BIT_LLSR_JALTS){
		DEBUG_EVENT("%s: Jitter Attenuator Limit Trip Set!\n",
					fe->name);
	}
	if (istatus & BIT_LLSR_OCC){
		DEBUG_EVENT("%s: Open Circuit is cleared!\n",
					fe->name);
	}
	if (istatus & BIT_LLSR_OCD){
		DEBUG_EVENT("%s: Open Circuit is detected!\n",
					fe->name);
	}
	if (istatus & BIT_LLSR_SCC){
		DEBUG_EVENT("%s: Short Circuit is cleared!\n",
					fe->name);
	}
	if (istatus & BIT_LLSR_SCD){
		DEBUG_EVENT("%s: Short Circuit is detected!\n",
					fe->name);
	}
	if (istatus & BIT_LLSR_LOSC){
		DEBUG_EVENT("%s: Lost of Signal is cleared!\n",
					fe->name);
	}
	if (istatus & BIT_LLSR_LOSD){
		DEBUG_EVENT("%s: Lost of Signal is detected!\n",
					fe->name);
	}
	WRITE_REG(REG_LLSR, 0xFF);
	return 0;
}

static int sdla_8te1_intr(sdla_fe_t *fe) 
{
	unsigned char	framer_istatus;
	unsigned char	liu_istatus;
	unsigned char	bert_istatus;
	
	WAN_ASSERT(fe->write_fe_reg == NULL);
	WAN_ASSERT(fe->read_fe_reg == NULL);

	framer_istatus = READ_REG(REG_GFISR);
	liu_istatus = READ_REG(REG_GLISR);
	bert_istatus = READ_REG(REG_GBISR);

	if (framer_istatus & (1 << WAN_FE_LINENO(fe))){
		sdla_8te1_framer_rx_intr(fe);
		sdla_8te1_framer_tx_intr(fe);
		WRITE_REG(REG_GFISR, (1<<WAN_FE_LINENO(fe)));
	}
	if (liu_istatus & (1 << WAN_FE_LINENO(fe))){
		sdla_8te1_liu_intr(fe);
		WRITE_REG(REG_GLISR, (1<<WAN_FE_LINENO(fe)));
	}
	if (bert_istatus & (1 << WAN_FE_LINENO(fe))){
		sdla_8te1_bert_intr(fe);
		WRITE_REG(REG_GBISR, (1<<WAN_FE_LINENO(fe)));
	}

	DEBUG_TEST("%s: FE Interrupt Alarms=0x%lX\n",
			fe->name,fe->fe_alarm);

	sdla_8te1_set_status(fe, fe->fe_alarm);

	return 0;
}

/*
 ******************************************************************************
 *				sdla_8te1_timer()	
 *
 * Description:
 * Arguments:
 * Returns:
 ******************************************************************************
 */
#if defined(__NetBSD__) || defined(__FreeBSD__) || defined(__OpenBSD__)
static void sdla_8te1_timer(void* pfe)
#elif defined(__WINDOWS__)
static void sdla_8te1_timer(IN PKDPC Dpc, void* pfe, void* arg2, void* arg3)
#else
static void sdla_8te1_timer(unsigned long pfe)
#endif
{
	sdla_fe_t	*fe = (sdla_fe_t*)pfe;
	sdla_t 		*card = (sdla_t*)fe->card;
	wan_device_t	*wandev = &card->wandev;

	if (wan_test_bit(TE_TIMER_KILL,(void*)&fe->fe_param.te_param.critical)){
		wan_clear_bit(TE_TIMER_RUNNING,(void*)&fe->fe_param.te_param.critical);
		return;
	}
	/*WAN_ASSERT1(wandev->te_enable_timer == NULL); */
	DEBUG_TIMER("%s: TE1 timer!\n", fe->name);
	/* Enable hardware interrupt for TE1 */

	if (wandev->fe_enable_timer){
		wandev->fe_enable_timer(fe->card);
	}else{
		sdla_8te1_polling(fe);
	}

	return;
}

/*
 ******************************************************************************
 *				sdla_8te1_enable_timer()	
 *
 * Description: Enable software timer interrupt in delay ms.
 * Arguments:
 * Returns:
 ******************************************************************************
 */
static void sdla_8te1_enable_timer(sdla_fe_t* fe, unsigned char cmd, unsigned long delay)
{
	sdla_t	*card = (sdla_t*)fe->card;

	WAN_ASSERT1(card == NULL);
	DEBUG_TEST("%s: %s:%d Cmd=0x%X\n",
			fe->name,__FUNCTION__,__LINE__,cmd);

#if defined (__WINDOWS__)
	if(KeGetCurrentIrql() > DISPATCH_LEVEL){
		/*	May get here on AFT card because front end interrupt
			is handled inside ISR not in DPC as on S514.
			The KeSetTimer() function is illegal inside ISR,
			so queue 'front_end_dpc_obj' DPC and this routine
			will be called again from xilinx_front_end_dpc().
		*/
		card->xilinx_fe_dpc.te_timer_delay = delay;
		fe->fe_param.te_param.timer_cmd=(unsigned char)cmd;

		if(KeInsertQueueDpc(&card->front_end_dpc_obj, NULL,
			(PVOID)ENABLE_TE_TIMER) == FALSE){

			DEBUG_TE1("Failed to queue 'front_end_dpc_obj'!\n");
		}else{
			DEBUG_TEST("Successfully queued 'front_end_dpc_obj'.\n");
		}
		return;
	}/* if() */
#endif
	if (wan_test_bit(TE_TIMER_KILL,(void*)&fe->fe_param.te_param.critical)){
		wan_clear_bit(
				TE_TIMER_RUNNING,
				(void*)&fe->fe_param.te_param.critical);
		return;
	}
	
	if (wan_test_bit(TE_TIMER_RUNNING,(void*)&fe->fe_param.te_param.critical)){
		if (fe->fe_param.te_param.timer_cmd == cmd){
			/* Just ignore current request */
			return;
		}
		DEBUG_TEST("%s: TE_TIMER_RUNNING: new_cmd=%X curr_cmd=%X\n",
					fe->name,
					cmd,
					fe->fe_param.te_param.timer_cmd);
		return;
	}

	wan_set_bit(TE_TIMER_RUNNING,(void*)&fe->fe_param.te_param.critical);
	
	fe->fe_param.te_param.timer_cmd=cmd;

	wan_add_timer(&fe->fe_param.te_param.timer, delay * HZ / 1000);
	return;	
}

/*
 ******************************************************************************
 *				sdla_8te1_polling()	
 *
 * Description:
 * Arguments:
 * Returns:
 ******************************************************************************
 */
static int sdla_8te1_polling(sdla_fe_t* fe)
{
	sdla_t	*card = (sdla_t*)fe->card;

	WAN_ASSERT(fe->write_fe_reg == NULL);
	WAN_ASSERT(fe->read_fe_reg == NULL);

	DEBUG_EVENT("ADBG: %s: TE1 Polling State=%s Cmd=0x%X!\n", 
			fe->name, fe->fe_status==FE_CONNECTED?"Con":"Disconn",
			fe->fe_param.te_param.timer_cmd);

	wan_clear_bit(TE_TIMER_RUNNING,(void*)&fe->fe_param.te_param.critical);
	
	switch(fe->fe_param.te_param.timer_cmd){
	case TE_LINKDOWN_TIMER:
		sdla_8te1_read_alarms(fe, 1);
	
		sdla_8te1_set_status(fe, fe->fe_alarm);
		if (fe->fe_status == FE_CONNECTED){
			sdla_8te1_enable_timer(fe, TE_LINKUP_TIMER, POLLING_TE1_TIMER);
		}else{
			sdla_8te1_enable_timer(fe, TE_LINKDOWN_TIMER, POLLING_TE1_TIMER);
		}
		break;

	case TE_LINKUP_TIMER:
		/* ALEX: 
		** Do not update protocol front end state from TE_LINKDOWN_TIMER
		** because it cause to stay longer in interrupt handler
	        ** (critical for XILINX code) */
		if (fe->fe_status == FE_CONNECTED){
			if (card->wandev.te_link_state){
				card->wandev.te_link_state(card);
			}
		}else{
			sdla_8te1_enable_timer(fe, TE_LINKDOWN_TIMER, POLLING_TE1_TIMER);
		}
		break;

	default:
		DEBUG_EVENT("%s: T1/E1 Polling cmd %X (Internal Error %s:%d)\n",
				fe->name, fe->fe_param.te_param.timer_cmd,
				__FUNCTION__,__LINE__);
		break;
	}
	return 0;			
}

/*
 ******************************************************************************
 *				sdla_8te1_flush_pmon()	
 *
 * Description: Flush Dallas performance monitoring counters
 * Arguments:
 * Returns:
 ******************************************************************************
 */
static int sdla_8te1_flush_pmon(sdla_fe_t *fe)
{
	fe->fe_pmon.u.te_pmon.pmon1 = 0;
	fe->fe_pmon.u.te_pmon.pmon2 = 0;
	fe->fe_pmon.u.te_pmon.pmon3 = 0;
	fe->fe_pmon.u.te_pmon.pmon4 = 0;
	return 0;
}

/*
 ******************************************************************************
 *				sdla_8te1_pmon()	
 *
 * Description: Read DLS performance monitoring counters
 * Arguments:
 * Returns:
 ******************************************************************************
 */
static int sdla_8te1_pmon(sdla_fe_t *fe)
{
	sdla_te_pmon_t	*pmon = &fe->fe_pmon.u.te_pmon;

	WAN_ASSERT(fe->write_fe_reg == NULL);
	WAN_ASSERT(fe->read_fe_reg == NULL);

	/* Line code violation count register */	
	pmon->lcv += READ_REG(REG_LCVCR1) << 8 | READ_REG(REG_LCVCR2);

	/* Framing bit for E1/T1 */
	pmon->frm_bit_error += READ_REG(REG_PCVCR1) << 8 | READ_REG(REG_PCVCR2);

	/* OOF Error for T1 or Far End Block Error for E1 */
	pmon->oof_errors += READ_REG(REG_FOSCR1) << 8 | READ_REG(REG_FOSCR2);

	return 0;
}


/*
 ******************************************************************************
 *				sdla_8te1_liu_alb()	
 *
 * Description:
 * Arguments:
 * Returns:
 ******************************************************************************
 */
static int sdla_8te1_liu_alb(sdla_fe_t* fe, unsigned char mode) 
{
	unsigned char	value;

	WAN_ASSERT(fe->write_fe_reg == NULL);
	WAN_ASSERT(fe->read_fe_reg == NULL);

	value = READ_REG(REG_LMCR);
	if (mode == WAN_TE1_ACTIVATE_LB){
		value |= BIT_LMCR_ALB;
	}else{
		value &= ~BIT_LMCR_ALB;
	}
	WRITE_REG(REG_LMCR, value);
	return 0;
}

/*
 ******************************************************************************
 *				sdla_8te1_liu_llb()	
 *
 * Description:
 * Arguments:
 * Returns:
 ******************************************************************************
 */
static int sdla_8te1_liu_llb(sdla_fe_t* fe, unsigned char mode) 
{
	unsigned char	value;

	WAN_ASSERT(fe->write_fe_reg == NULL);
	WAN_ASSERT(fe->read_fe_reg == NULL);

	value = READ_REG(REG_LMCR);
	if (mode == WAN_TE1_ACTIVATE_LB){
		value |= BIT_LMCR_LLB;
	}else{
		value &= ~BIT_LMCR_LLB;
	}
	WRITE_REG(REG_LMCR, value);
	return 0;
}

/*
 ******************************************************************************
 *				sdla_8te1_liu_rlb()	
 *
 * Description:
 * Arguments:
 * Returns:
 ******************************************************************************
 */
static int sdla_8te1_liu_rlb(sdla_fe_t* fe, unsigned char mode) 
{
	unsigned char	value;

	WAN_ASSERT(fe->write_fe_reg == NULL);
	WAN_ASSERT(fe->read_fe_reg == NULL);

	value = READ_REG(REG_LMCR);
	if (mode == WAN_TE1_ACTIVATE_LB){
		value |= BIT_LMCR_RLB;
	}else{
		value &= ~BIT_LMCR_RLB;
	}
	WRITE_REG(REG_LMCR, value);
	return 0;
}

/*
 ******************************************************************************
 *				sdla_8te1_fr_flb()	
 *
 * Description:
 * Arguments:
 * Returns:
 ******************************************************************************
 */
static int sdla_8te1_fr_flb(sdla_fe_t* fe, unsigned char mode) 
{
	unsigned char	value;

	WAN_ASSERT(fe->write_fe_reg == NULL);
	WAN_ASSERT(fe->read_fe_reg == NULL);

	value = READ_REG(REG_RCR3);
	if (mode == WAN_TE1_ACTIVATE_LB){
		value |= BIT_RCR3_FLB;
	}else{
		value &= ~BIT_RCR3_FLB;
	}
	WRITE_REG(REG_LMCR, value);
	return 0;
}

/*
 ******************************************************************************
 *				sdla_8te1_fr_plb()
 *
 * Description:
 * Arguments:
 * Returns:
 ******************************************************************************
 */
static int sdla_8te1_fr_plb(sdla_fe_t* fe, unsigned char mode) 
{
	unsigned char	value;

	WAN_ASSERT(fe->write_fe_reg == NULL);
	WAN_ASSERT(fe->read_fe_reg == NULL);

	value = READ_REG(REG_RCR3);
	if (mode == WAN_TE1_ACTIVATE_LB){
		value |= BIT_RCR3_PLB;
	}else{
		value &= ~BIT_RCR3_PLB;
	}
	WRITE_REG(REG_LMCR, value);
	return 0;
}

/*
 ******************************************************************************
 *				sdla_8te1_set_lbmode()	
 *
 * Description:
 * Arguments:
 * Returns:
 ******************************************************************************
 */
static int 
sdla_8te1_set_lbmode(sdla_fe_t *fe, unsigned char type, unsigned char mode) 
{

	WAN_ASSERT(fe->write_fe_reg == NULL);
	WAN_ASSERT(fe->read_fe_reg == NULL);
	DEBUG_EVENT("%s: %s %s mode...\n",
			fe->name,
			WAN_TE1_LB_MODE_DECODE(mode),
			WAN_TE1_LB_TYPE_DECODE(type));
	switch(type){
	case WAN_TE1_LIU_ALB_MODE:
		sdla_8te1_liu_alb(fe, mode);
		break;
	case WAN_TE1_LIU_LLB_MODE:
		sdla_8te1_liu_llb(fe, mode);
		break;
	case WAN_TE1_LIU_RLB_MODE:
		sdla_8te1_liu_rlb(fe, mode);
		break;
	case WAN_TE1_LIU_DLB_MODE:
		sdla_8te1_liu_llb(fe, mode);
		sdla_8te1_liu_rlb(fe, mode);
		break;
	case WAN_TE1_FR_FLB_MODE:
		sdla_8te1_fr_flb(fe, mode);
		break;
	case WAN_TE1_FR_PLB_MODE:
		sdla_8te1_fr_plb(fe, mode);
		break;
	case WAN_TE1_FR_RLB_MODE:
		break;
	}
	return 0;
}

/*
 ******************************************************************************
 *				sdla_8te1_udp()	
 *
 * Description:
 * Arguments:
 * Returns:
 ******************************************************************************
 */
static int sdla_8te1_udp(sdla_fe_t *fe, void* p_udp_cmd, unsigned char* data)
{
	wan_cmd_t	*udp_cmd = (wan_cmd_t*)p_udp_cmd;
	int		err = 0;

	switch(udp_cmd->wan_cmd_command){
	case WAN_GET_MEDIA_TYPE:
		data[0] = (IS_T1_FEMEDIA(fe) ? WAN_MEDIA_T1 :
			   IS_E1_FEMEDIA(fe) ? WAN_MEDIA_E1 :
					    WAN_MEDIA_NONE);
		udp_cmd->wan_cmd_return_code = WAN_CMD_OK;
		udp_cmd->wan_cmd_data_len = sizeof(unsigned char); 
		break;

	case WAN_FE_SET_LB_MODE:
		/* Activate/Deactivate Line Loopback modes */
	    	err = sdla_8te1_set_lbmode(fe, data[0], data[1]); 
	    	udp_cmd->wan_cmd_return_code = 
				(!err) ? WAN_CMD_OK : WAN_UDP_FAILED_CMD;
	    	udp_cmd->wan_cmd_data_len = 0x00;
		break;

	case WAN_FE_GET_STAT:
 	        /* TE1_56K Read T1/E1/56K alarms */
	  	*(unsigned long *)&data[0] = fe->fe_alarm;
		/* TE1 Update T1/E1 perfomance counters */
		sdla_8te1_pmon(fe);
	        memcpy(&data[sizeof(unsigned long)],
			&fe->fe_pmon.u.te_pmon,
			sizeof(sdla_te_pmon_t));
	        udp_cmd->wan_cmd_return_code = WAN_CMD_OK;
	    	udp_cmd->wan_cmd_data_len = 
			sizeof(unsigned long) + sizeof(sdla_te_pmon_t); 

		break;

 	case WAN_FE_FLUSH_PMON:
		/* TE1 Flush T1/E1 pmon counters */
		sdla_8te1_flush_pmon(fe);
	        udp_cmd->wan_cmd_return_code = WAN_CMD_OK;
		break;
 
	case WAN_FE_GET_CFG:
		/* Read T1/E1 configuration */
	       	memcpy(&data[0],
	    		&fe->fe_cfg,
		      	sizeof(sdla_fe_cfg_t));
	    	udp_cmd->wan_cmd_return_code = WAN_CMD_OK;
    	    	udp_cmd->wan_cmd_data_len = sizeof(sdla_fe_cfg_t);
	    	break;

	case WAN_FE_SET_DEBUG_MODE:
#if 0
		switch(data[0]){
		case WAN_FE_DEBUG_RBS:
			if (data[1] & WAN_FE_DEBUG_RBS_READ){
				DEBUG_EVENT("%s: Reading RBS status!\n",
					fe->name);
#if 1
				sdla_te_enable_timer(fe, TE_RBS_READ, POLLING_TE1_TIMER);
#else
				sdla_te_rbs_print(fe, 0);
#endif
	    			udp_cmd->wan_cmd_return_code = WAN_CMD_OK;
			}else if (data[1] & WAN_FE_DEBUG_RBS_RX_ENABLE){
				/* Enable extra debugging */
				DEBUG_EVENT("%s: Enable RBS RX DEBUG mode!\n",
					fe->name);
				fe->fe_debug |= WAN_FE_DEBUG_RBS_RX_ENABLE;
				sdla_te_rbs_print(fe, 1);
	    			udp_cmd->wan_cmd_return_code = WAN_CMD_OK;
			}else if (data[1] & WAN_FE_DEBUG_RBS_TX_ENABLE){
				/* Enable extra debugging */
				DEBUG_EVENT("%s: Enable RBS TX DEBUG mode!\n",
					fe->name);
				fe->fe_debug |= WAN_FE_DEBUG_RBS_TX_ENABLE;
				sdla_te_rbs_print(fe, 1);
	    			udp_cmd->wan_cmd_return_code = WAN_CMD_OK;
			}else if (data[1] & WAN_FE_DEBUG_RBS_RX_DISABLE){
				/* Disable extra debugging */
				DEBUG_EVENT("%s: Disable RBS RX DEBUG mode!\n",
					fe->name);
				fe->fe_debug &= ~WAN_FE_DEBUG_RBS_RX_ENABLE;
				sdla_te_rbs_print(fe, 1);
	    			udp_cmd->wan_cmd_return_code = WAN_CMD_OK;
			}else if (data[1] & WAN_FE_DEBUG_RBS_TX_DISABLE){
				/* Disable extra debugging */
				DEBUG_EVENT("%s: Disable RBS TX DEBUG mode!\n",
					fe->name);
				fe->fe_debug &= ~WAN_FE_DEBUG_RBS_TX_ENABLE;
				sdla_te_rbs_print(fe, 1);
	    			udp_cmd->wan_cmd_return_code = WAN_CMD_OK;
			}
			break;
		case WAN_FE_DEBUG_ALARM:
			if (data[1] & WAN_FE_DEBUG_ALARM_AIS_ENABLE){
				/* Enable TX AIS alarm */
				DEBUG_EVENT("%s: Enable TX AIS alarm!\n",
					fe->name);
				WRITE_REG(REG_E1_TRAN_TX_ALARM_CTRL,
					READ_REG(REG_E1_TRAN_TX_ALARM_CTRL) |
						BIT_E1_TRAN_TX_ALARM_AIS); 
				WRITE_REG(REG_TX_LINE_CONF,
					READ_REG(REG_TX_LINE_CONF) |
						BIT_TX_LINE_CONF_TAISEN);
	    			udp_cmd->wan_cmd_return_code = WAN_CMD_OK;
			}else if (data[1] & WAN_FE_DEBUG_ALARM_AIS_DISABLE){
				/* Disable TX AIS alarm */
				DEBUG_EVENT("%s: Disable TX AIS alarm!\n",
					fe->name);
				WRITE_REG(REG_E1_TRAN_TX_ALARM_CTRL,
					READ_REG(REG_E1_TRAN_TX_ALARM_CTRL) &
						~BIT_E1_TRAN_TX_ALARM_AIS); 
				WRITE_REG(REG_TX_LINE_CONF,
					READ_REG(REG_TX_LINE_CONF) &
						~BIT_TX_LINE_CONF_TAISEN);
	    			udp_cmd->wan_cmd_return_code = WAN_CMD_OK;
			}
			break;
		default:
			udp_cmd->wan_cmd_return_code = WAN_UDP_INVALID_CMD;
			break;
		}
    	    	udp_cmd->wan_cmd_data_len = 0;
		break;
	case WAN_FE_TX_MODE:
		switch(data[0]){
		case WAN_FE_TXMODE_ENABLE:
			DEBUG_EVENT("%s: Enable Transmitter!\n",
					fe->name);
			WRITE_REG(REG_XLPG_LINE_CFG, fe->fe_param.te_param.xlpg_scale);
			udp_cmd->wan_cmd_return_code = WAN_CMD_OK;
			break;
		case WAN_FE_TXMODE_DISABLE:
			DEBUG_EVENT("%s: Disable Transmitter (tx tri-state mode)!\n",
					fe->name);
			WRITE_REG(REG_XLPG_LINE_CFG, BIT_XLPG_LINE_CFG_HIGHZ);
			udp_cmd->wan_cmd_return_code = WAN_CMD_OK;
			break;
		}
		break;
#endif
	default:
		udp_cmd->wan_cmd_return_code = WAN_UDP_INVALID_CMD;
	    	udp_cmd->wan_cmd_data_len = 0;
		break;
	}
	return 0;
}

static int 
sdla_8te1_update_alarm_info(sdla_fe_t* fe, struct seq_file* m, int* stop_cnt)
{

#if !defined(__WINDOWS__)
	PROC_ADD_LINE(m,
		"=============================== %s Alarms ===============================\n",
		FE_MEDIA_DECODE(fe));
	PROC_ADD_LINE(m,
		PROC_STATS_ALARM_FORMAT,
		"ALOS", WAN_TE_ALOS_ALARM(fe->fe_alarm), 
		"LOS", WAN_TE_LOS_ALARM(fe->fe_alarm));
	PROC_ADD_LINE(m,
		PROC_STATS_ALARM_FORMAT,
		"RED", WAN_TE_RED_ALARM(fe->fe_alarm), 
		"AIS", WAN_TE_AIS_ALARM(fe->fe_alarm));
	if (IS_T1_FEMEDIA(fe)){
		PROC_ADD_LINE(m,
			 PROC_STATS_ALARM_FORMAT,
			 "YEL", WAN_TE_YEL_ALARM(fe->fe_alarm),
			 "OOF", WAN_TE_OOF_ALARM(fe->fe_alarm));
	}else{ 
		PROC_ADD_LINE(m,
			PROC_STATS_ALARM_FORMAT,
			"OOF", WAN_TE_OOF_ALARM(fe->fe_alarm), 
			"", "");
	}
	return m->count;	
#endif
	return 0;
}

static int 
sdla_8te1_update_pmon_info(sdla_fe_t* fe, struct seq_file* m, int* stop_cnt)
{
		
#if !defined(__WINDOWS__)
	PROC_ADD_LINE(m,
		 "=========================== %s PMON counters ============================\n",
		 FE_MEDIA_DECODE(fe));
	PROC_ADD_LINE(m,
		PROC_STATS_PMON_FORMAT,
		"Framing Bit Error", fe->fe_pmon.u.te_pmon.frm_bit_error,
		"Line Code Violation", fe->fe_pmon.u.te_pmon.lcv);
	if (IS_T1_FEMEDIA(fe)){
		PROC_ADD_LINE(m,
			PROC_STATS_PMON_FORMAT,
			"Out of Frame Errors", fe->fe_pmon.u.te_pmon.bit_errors,
			"Bit Errors", fe->fe_pmon.u.te_pmon.bit_errors);
	}else{ 
		PROC_ADD_LINE(m,
			PROC_STATS_PMON_FORMAT,
			"Far End Block Errors", fe->fe_pmon.u.te_pmon.far_end_blk_errors,
			"CRC Errors", fe->fe_pmon.u.te_pmon.crc_errors);
	}
	return m->count;
#endif
	return 0;
}


